"""Implementation of the Federated Averaging with momentum [FedAVGM19]_ algorithm.

References:
    .. [FedAVGM19] Tzu-Ming Harry Hsu, Hang Qi, Matthew Brown. Measuring the Effects of
       Non-Identical Data Distribution for Federated Visual Classification. In arXiv (2019).
       URL: https://arxiv.org/abs/1909.06335
"""
import sys
from copy import deepcopy
from typing import Any, Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils.model import state_dict_zero_like  # NOQA

__all__ = [
    "FedAVGMServer",
    "FedAVGM"
]


class FedAVGMServer(Server):
    """Server class for the FedAVGM algorithm.

    Args:
        model (Module): The model to be trained.
        test_set (FastDataLoader): The test data.
        clients (Iterable[Client]): The clients participating in the federated learning process.
        eval_every (int, optional): Evaluate the model every `eval_every` rounds. Defaults to 1.
        weighted (bool, optional): Use weighted averaging. Defaults to True.
        momentum (float, optional): The momentum hyper-parameter. Defaults to 0.9.
    """

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = True,
                 momentum: float = 0.9,
                 **kwargs: dict[str, Any]):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.hyper_params.update(momentum=momentum)
        self.momentum_vector = None

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        prev_model_sd = deepcopy(self.model.state_dict())
        super().aggregate(eligible, client_models)
        agg_model_sd = self.model.state_dict()

        if self.momentum_vector is None:
            self.momentum_vector = state_dict_zero_like(prev_model_sd)
        else:
            for key in prev_model_sd:
                self.momentum_vector[key].data = self.hyper_params.momentum * \
                    self.momentum_vector[key].data + \
                    prev_model_sd[key].data - agg_model_sd[key].data

        for key in prev_model_sd:
            agg_model_sd[key].data = prev_model_sd[key].data - self.momentum_vector[key].data

        self.model.load_state_dict(agg_model_sd)


class FedAVGM(CentralizedFL):

    def get_server_class(self) -> Server:
        return FedAVGMServer



"""Implementation of the [LG-FedAVG20]_ algorithm.

References:
    .. [LG-FedAVG20] Paul Pu Liang, Terrance Liu, Liu Ziyin, Nicholas B. Allen, Randy P. Auerbach,
       David Brent, Ruslan Salakhutdinov, Louis-Philippe Morency. Think Locally, Act Globally:
       Federated Learning with Local and Global Representations. In arXiv (2020).
       URL: https://arxiv.org/abs/2001.01523
"""
import sys
from typing import Any, Iterable

# from torch.nn import CrossEntropyLoss
from torch.nn.modules import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderHeadNet, HeadGlobalEncoderLocalNet  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, get_model  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA

__all__ = [
    "LGFedAVGClient",
    "LGFedAVGServer",
    "LGFedAVG"
]

# The implementation is almost identical to FedPerClient
# The difference lies in the part of the network that is local and the part that is global.
# One is the inverse of the other.


class LGFedAVGClient(Client):

    def __init__(self,
                 index: int,
                 model: EncoderHeadNet,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,  # In the paper it is fixed to CrossEntropyLoss
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set,
                         test_set=test_set, optimizer_cfg=optimizer_cfg, loss_fn=loss_fn,
                         local_epochs=local_epochs, fine_tuning_epochs=fine_tuning_epochs, **kwargs)
        self._load_from_cache()
        self.model = HeadGlobalEncoderLocalNet(model)
        self._save_to_cache()

    def send_model(self):
        self.channel.send(Message(self.model.get_global(), "model", self, inmemory=True),
                          self.server)

    def receive_model(self) -> None:
        msg = self.channel.receive(self, self.server, msg_type="model")
        safe_load_state_dict(self.model.get_global(), msg.payload.state_dict())


class LGFedAVGServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,  # not used
                 clients: Iterable[Client],
                 weighted: bool = False):
        super().__init__(model=model, test_set=None, clients=clients, weighted=weighted)


class LGFedAVG(PersonalizedFL):

    def get_client_class(self) -> Client:
        return LGFedAVGClient

    def get_server_class(self) -> Server:
        return LGFedAVGServer



"""
The module :mod:`fluke.client` provides the base classes for the clients in :mod:`fluke`.
"""
from __future__ import annotations

import sys
from typing import Any, Literal

import torch
from torch import device
from torch.nn import Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

sys.path.append(".")

from . import DDict, FlukeCache, FlukeENV, ObserverSubject  # NOQA
from .comm import Channel, Message  # NOQA
from .data import FastDataLoader  # NOQA
from .evaluation import Evaluator  # NOQA
from .server import Server  # NOQA
from .utils import (OptimizerConfigurator, cache_obj, clear_cuda_cache,  # NOQA
                    retrieve_obj)
from .utils.model import ModOpt, safe_load_state_dict  # NOQA

__all__ = [
    "Client",
    "PFLClient"
]


class Client(ObserverSubject):
    """Base :class:`Client` class. This class is the base class for all clients in :mod:`fluke`.
    The standard behavior of a client is based on the Federated Averaging algorithm.
    The default behavior of a client includes:

    - Receiving the global model from the server;
    - Training the model locally for a number of epochs using the local training set;
    - Sending the updated local model back to the server;
    - (Optional) Evaluating the model on the local test set before and after the training.

    Attributes:
        hyper_params (DDict): The hyper-parameters of the client. The default hyper-parameters are:

            - ``loss_fn``: The loss function.
            - ``local_epochs``: The number of local epochs.
            - ``fine_tuning_epochs``: The number of fine-tuning epochs, i.e., the number of epochs
              to train the model after the federated learning process.
            - ``clipping``: The clipping value for the gradients.

            When a new client class inherits from this class, it must add all its hyper-parameters
            to this dictionary.

        train_set (FastDataLoader): The local training set.
        test_set (FastDataLoader): The local test set.
        optimizer_cfg (OptimizerConfigurator): The optimizer configurator. This is used to create
          the optimizer and the learning rate scheduler.
        device (torch.device): The device where the client trains the model. By default, it is the
          device defined in :class:`fluke.FlukeENV`.

    Attention:
        **The client should not directly call methods of the server**. The communication between the
        client and the server must be done through the :attr:`channel`.

    Important:
        When inheriting from this class, make sure to handle the caching of the model, optimizer,
        scheduler and any other attribute that should be cached. The caching is done automatically
        by the methods :meth:`_load_from_cache` and :meth:`_save_to_cache`. The method
        :meth:`_load_from_cache` is called before the local update and the method
        :meth:`_save_to_cache` is called after the local update. If the client has additional
        attributes that should be cached, add them to the ``_attr_to_cache`` list.
        If the client has an additional model, optimizer, and scheduler, these object should be
        handled via a :class:`fluke.utils.model.ModOpt` private object. The
        :class:`fluke.utils.model.ModOpt` object is used to store the model, optimizer, and
        scheduler. To access the model, optimizer, and scheduler, define the
        corresponding properties (e.g., :attr:`model`, :attr:`optimizer`, :attr:`scheduler`).

    Caution:
        When inheriting from this class, make sure to put all the specific hyper-parameters in the
        :attr:`hyper_params` attribute. In this way ``fluke`` can properly handle the
        hyper-parameters of the client in the federated learning process.

        For example:

        .. code-block:: python
            :linenos:

            class MyClient(Client):
                # We omit the type hints for brevity
                def __init__(self, index, ..., my_param):
                    super().__init__(index, ...)
                    self.hyper_params.update(my_param=my_param) # This is important
    """

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__()
        self.train_set: FastDataLoader = train_set
        self.test_set: FastDataLoader = test_set
        self.device: device = FlukeENV().get_device()
        self.hyper_params: DDict = DDict(
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping
        )

        self._index: int = index
        self._modopt: ModOpt = ModOpt()
        self._optimizer_cfg: OptimizerConfigurator = optimizer_cfg
        self._server: Server = None
        self._channel: Channel = None
        self._last_round: int = 0
        # List of additional attributes to cache
        self._attr_to_cache: list[str] = []

    @property
    def model(self) -> Module:
        """The client's local model.

        Warning:
            If the model is stored in the cache, the method retrieves it from the cache but does not
            remove it. Thus, the performance may be affected if this property is used to get the
            model multiple times while the model is in the cache.

        Returns:
            torch.nn.Module: The local model.
        """
        if isinstance(self._modopt, FlukeCache._ObjectRef):
            return retrieve_obj("_modopt", self, pop=False).model
        return self._modopt.model

    @model.setter
    def model(self, model: Module) -> None:
        self._modopt.model = model

    @property
    def optimizer(self) -> Optimizer:
        """The optimizer of the client.

        Warning:
            If the optimizer is stored in the cache, the method retrieves it from the cache but does
            not remove it. Thus, the performance may be affected if this property is used to get the
            optimizer multiple times while the optimizer is in the cache.

        Returns:
            torch.optim.Optimizer: The optimizer.
        """
        if isinstance(self._modopt, FlukeCache._ObjectRef):
            return retrieve_obj("_modopt", self, pop=False).optimizer
        return self._modopt.optimizer

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer) -> None:
        self._modopt.optimizer = optimizer

    @property
    def scheduler(self) -> LRScheduler:
        """The learning rate scheduler of the client.

        Warning:
            If the scheduler is stored in the cache, the method retrieves it from the cache but does
            not remove it. Thus, the performance may be affected if this property is used to get the
            scheduler multiple times while the scheduler is in the cache.

        Returns:
            torch.optim.lr_scheduler.LRScheduler: The learning rate scheduler.
        """
        if isinstance(self._modopt, FlukeCache._ObjectRef):
            return retrieve_obj("_modopt", self, pop=False).scheduler
        return self._modopt.scheduler

    @scheduler.setter
    def scheduler(self, scheduler: LRScheduler) -> None:
        self._modopt.scheduler = scheduler

    @property
    def index(self) -> int:
        """The client identifier. This might be useful to identify the client in the federated
        learning process for logging or debugging purposes.

        Returns:
            int: The client identifier.
        """
        return self._index

    @property
    def local_model(self) -> Module:
        """The client's local model.
        This is an alias for `model`.

        Returns:
            torch.nn.Module: The local model.
        """
        return self.model

    @property
    def n_examples(self) -> int:
        """The number of examples in the local training set.

        Returns:
            int: The number of examples in the local training set.
        """
        if isinstance(self.train_set, FastDataLoader):
            return self.train_set.size
        else:
            # Assume that the train_set is a PyTorch DataLoader
            return len(self.train_set.dataset)

    @property
    def channel(self) -> Channel:
        """The communication channel.

        Attention:
            Use this channel to exchange data/information with the server.

        Returns:
            Channel: The communication channel.
        """
        return self._channel

    @property
    def server(self) -> Server:
        """The server to which the client is connected. This reference must only be
        used to send messages through the channel.

        Returns:
            Server: The server.
        """
        return self._server

    def set_server(self, server: Server) -> None:
        """Set the reference to the server. Along with the server, the communication channel is also
        set and the client must use this channel to communicate with the server.

        Args:
            server (Server): The server that orchestrates the federated learning process.
        """
        self._server = server
        self._channel = server.channel

    def receive_model(self) -> None:
        """Receive the global model from the server. This method is responsible for receiving the
        global model from the server and updating the local model accordingly. The model is received
        as a payload of a :class:`fluke.comm.Message` with  ``msg_type="model"`` from the inbox
        of the client itself. The method uses the channel to receive the message.
        """
        msg = self.channel.receive(self, self.server, msg_type="model")
        if self.model is None:
            self.model = msg.payload
        else:
            safe_load_state_dict(self.model, msg.payload.state_dict())

    def send_model(self) -> None:
        """Send the current model to the server. The model is sent as a :class:`fluke.comm.Message`
        with ``msg_type`` "model" to the server. The method uses the channel to send the message.
        """
        self.channel.send(Message(self.model, "model", self, inmemory=True), self.server)

    def local_update(self, current_round: int) -> None:
        """Client's local update procedure.
        Before starting the local training, the client receives the global model from the server.
        Then, the training occurs for a number of ``hyper_params.local_epochs`` epochs using the
        local training set and as loss function the one defined in ``hyper_params.loss_fn``.
        The training happens on the device defined in the client. After the training, the client
        sends the model to the server.

        Args:
            current_round (int): The current round of the federated learning process.
        """
        self._last_round = current_round
        self._load_from_cache()
        self.receive_model()

        if FlukeENV().get_eval_cfg().pre_fit:
            metrics = self.evaluate(FlukeENV().get_evaluator(), self.test_set)
            if metrics:
                self._notify_evaluation(current_round, "pre-fit", metrics)

        self._notify_start_fit(current_round)
        loss = self.fit()
        self._notify_end_fit(current_round, loss)

        if FlukeENV().get_eval_cfg().post_fit:
            metrics = self.evaluate(FlukeENV().get_evaluator(), self.test_set)
            if metrics:
                self._notify_evaluation(current_round, "post-fit", metrics)

        self.send_model()
        self._save_to_cache()

    def _clip_grads(self, model: Module) -> None:
        if self.hyper_params.clipping > 0:
            torch.nn.utils.clip_grad_norm_(model.parameters(), self.hyper_params.clipping)

    def fit(self, override_local_epochs: int = 0) -> float:
        """Client's local training procedure.

        Args:
            override_local_epochs (int, optional): Overrides the number of local epochs,
                by default 0 (use the default number of local epochs as in
                ``hyper_params.local_epochs``).

        Returns:
            float: The average loss of the model during the training.
        """
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)

        self.model.train()
        self.model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        clear_cuda_cache()
        return running_loss

    def evaluate(self,
                 evaluator: Evaluator,
                 test_set: FastDataLoader) -> dict[str, float]:
        """Evaluate the local model on the client's :attr:`test_set`. If the test set is not set or
        the client has not received the global model from the server, the method returns an empty
        dictionary.

        Args:
            evaluator (Evaluator): The evaluator to use for the evaluation.
            test_set (FastDataLoader): The test set to use for the evaluation.

        Returns:
            dict[str, float]: The evaluation results. The keys are the metrics and the values are
            the results.
        """
        model = self.model  # ensure to call the retrieve_obj only once
        if test_set is not None and model is not None:
            evaluation = evaluator.evaluate(self._last_round, model, test_set,
                                            device=self.device)
            return evaluation
        return {}

    def finalize(self) -> None:
        """Finalize the client. This method is called at the end of the federated learning process.
        The default behavior is to receive the global model from the server that is then potentially
        used to evaluate the performance of the model on the local test set.

        Attention:
            When inheriting from this class, make sure to override this method if this behavior is
            not desired.
        """
        self._load_from_cache()
        # This only happens if the federation is stopped with a KeyboardInterrupt
        # and the client was doing the local update
        if self._modopt is None:
            return
        self.receive_model()

        if self.hyper_params.fine_tuning_epochs > 0:
            self.fit(self.hyper_params.fine_tuning_epochs)

        if FlukeENV().get_eval_cfg().pre_fit:
            metrics = self.evaluate(FlukeENV().get_evaluator(), self.test_set)
            if metrics:
                self._notify_evaluation(-1, "pre-fit", metrics)

        self._save_to_cache()

    def state_dict(self) -> dict[str, Any]:
        """Get the client state as a dictionary.

        Returns:
            dict: The client state.
        """
        modopt = retrieve_obj("_modopt", self, pop=False) \
            if isinstance(self._modopt, FlukeCache._ObjectRef) else self._modopt
        return {
            "modopt": modopt.state_dict(),
            "index": self.index,
            "last_round": self._last_round
        }

    def save(self, path: str) -> None:
        """Save the client state to a file.

        Args:
            path (str): The path to the file where the client state will be saved.
        """
        torch.save(self.state_dict(), path)

    def load(self, path: str, model: Module) -> dict[str, Any]:
        """Load the client state from a file.

        Args:
            path (str): The path to the file where the client state is saved.
            model (torch.nn.Module): The model to use for the client.

        Returns:
            dict: The loaded lient state.
        """
        state = torch.load(path, weights_only=True)
        if state["modopt"]["model"] is not None:
            self.model = model
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)
            self._modopt.load_state_dict(state["modopt"])
        else:
            self.model = None

        self._index = state["index"]
        self._last_round = state["last_round"]
        return state

    def __str__(self) -> str:
        hpstr = ", ".join([f"{h}={str(v)}" for h, v in self.hyper_params.items()])
        hpstr = ", " + hpstr if hpstr else ""
        return f"{self.__class__.__name__}[{self._index}](optim={self._optimizer_cfg}, " + \
            f"batch_size={self.train_set.batch_size}{hpstr})"

    def __repr__(self) -> str:
        return str(self)

    def _notify_evaluation(self,
                           round: int,
                           phase: Literal["pre-fit", "post-fit"],
                           metrics: dict[str, float]) -> None:
        for obs in self._observers:
            obs.client_evaluation(round, self.index, phase, metrics)

    def _notify_start_fit(self, round: int) -> None:
        for obs in self._observers:
            obs.start_fit(round, self.index, self.model)

    def _notify_end_fit(self, round: int, loss: float) -> None:
        for obs in self._observers:
            obs.end_fit(round, self.index, self.model, loss)

    def _load_from_cache(self) -> None:
        """Load the model, optimizer, and scheduler from the cache.
        The method retrieves the model, optimizer, and scheduler from the cache and sets them as
        the client's model, optimizer, and scheduler. The method should be called before the
        local update.

        Potential additional attributes that should be loaded from the cache should be added to the
        ``_attr_to_cache`` list.
        """
        if not FlukeENV().is_inmemory():
            for attr_name, attr_value in vars(self).items():
                if isinstance(attr_value, FlukeCache._ObjectRef):
                    new_value = retrieve_obj(attr_name, self)
                    setattr(self, attr_name, new_value)

    def _save_to_cache(self) -> None:
        """Save the model, optimizer, and scheduler to the cache.
        The method should be called after the local update.

        Potential additional attributes that should be saved to the cache should be added to the
        ``_attr_to_cache`` list.
        """
        if not FlukeENV().is_inmemory():
            # Cache the model, optimizer, and scheduler
            for attr_name, attr_value in vars(self).items():
                if not isinstance(attr_value, FlukeCache._ObjectRef) and \
                        isinstance(attr_value, ModOpt):
                    setattr(self, attr_name, cache_obj(attr_value, attr_name, self))

            # Cache additional attributes
            for attr in self._attr_to_cache:
                obj = getattr(self, attr)
                if not isinstance(obj, FlukeCache._ObjectRef):
                    setattr(self, attr, cache_obj(obj, attr, self))


class PFLClient(Client):
    """Personalized Federated Learning client.
    This class is a personalized version of the :class:`Client` class. It is used to implement
    personalized federated learning algorithms. The main difference is that the client has a
    personalized model (i.e., the attribute :attr:`personalized_model`).

    Note:
        The client evaluation is performed using :attr:`personalized_model` instead of the global
        model (i.e., :attr:`fluke.client.Client.model`).
    """

    def __init__(self,
                 index: int,
                 model: Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self._personalized_modopt: ModOpt = ModOpt(model=model)
        self._save_to_cache()

    @property
    def personalized_model(self) -> Module:
        """The personalized model.

        Warning:
            If the model is stored in the cache, the method retrieves it from the cache but does
            not remove it. Thus, the performance may be affected if this property is used to get the
            model multiple times while the model is in the cache.

        Returns:
            torch.nn.Module: The personalized model.
        """
        if isinstance(self._modopt, FlukeCache._ObjectRef):
            return retrieve_obj("_personalized_modopt", self, pop=False).model
        return self._personalized_modopt.model

    @personalized_model.setter
    def personalized_model(self, model: Module | FlukeCache._ObjectRef) -> None:
        self._personalized_modopt.model = model

    @property
    def pers_optimizer(self) -> Optimizer:
        """The optimizer of the personalized model.

        Warning:
            If the optimizer is stored in the cache, the method retrieves it from the cache but does
            not remove it. Thus, the performance may be affected if this property is used to get the
            optimizer multiple times while the optimizer is in the cache.

        Returns:
            torch.optim.Optimizer: The optimizer.
        """
        if isinstance(self._modopt, FlukeCache._ObjectRef):
            return retrieve_obj("_personalized_modopt", self, pop=False).optimizer
        return self._personalized_modopt.optimizer

    @pers_optimizer.setter
    def pers_optimizer(self, optimizer: Optimizer | FlukeCache._ObjectRef) -> None:
        self._personalized_modopt.optimizer = optimizer

    @property
    def pers_scheduler(self) -> LRScheduler:
        """The learning rate scheduler of the personalized model.

        Warning:
            If the scheduler is stored in the cache, the method retrieves it from the cache but does
            not remove it. Thus, the performance may be affected if this property is used to get the
            scheduler multiple times while the scheduler is in the cache.

        Returns:
            torch.optim.lr_scheduler.LRScheduler: The learning rate scheduler.
        """
        if isinstance(self._modopt, FlukeCache._ObjectRef):
            return retrieve_obj("_personalized_modopt", self, pop=False).scheduler
        return self._personalized_modopt.scheduler

    @pers_scheduler.setter
    def pers_scheduler(self, scheduler: LRScheduler) -> None:
        self._personalized_modopt.scheduler = scheduler

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        """Evaluate the personalized model on the :attr:`test_set`.
        If the test set is not set or the client has no personalized model, the method returns an
        empty dictionary.

        Args:
            evaluator (Evaluator): The evaluator to use for the evaluation.
            test_set (FastDataLoader): The test set to use for the evaluation.

        Returns:
            dict[str, float]: The evaluation results. The keys are the metrics and the values are
            the results.
        """
        if test_set is not None and self.personalized_model is not None:
            return evaluator.evaluate(self._last_round,
                                      self.personalized_model,
                                      test_set,
                                      device=self.device)
        return {}

    def state_dict(self) -> dict[str, Any]:
        state = super().state_dict()
        pmodopt = retrieve_obj("_personalized_modopt", self, pop=False) \
            if isinstance(self._personalized_modopt, FlukeCache._ObjectRef) \
            else self._personalized_modopt
        state["personalized_modopt"] = pmodopt.state_dict()
        return state

    def load(self, path: str, model: Module) -> dict[str, Any]:
        state = super().load(path, model)
        self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.personalized_model)
        self._personalized_modopt.load_state_dict(state["personalized_modopt"])
        return state

    @property
    def local_model(self) -> Module:
        """The client's local model.
        This is an alias for :attr:`personalized_model`.

        Returns:
            torch.nn.Module: The local model.
        """
        return self.personalized_model



"""This module contains the classes for the communication between the clients and the server."""
from __future__ import annotations

import sys
import uuid
import warnings
from collections import defaultdict
from copy import deepcopy
from typing import Any, Optional

import numpy as np
import torch

sys.path.append(".")

from . import ObserverSubject, FlukeENV, FlukeCache  # NOQA
from .utils import retrieve_obj, cache_obj  # NOQA

__all__ = [
    'Message',
    'Channel',
    'ChannelObserver'
]


class Message:
    """This class represents a message that can be exchanged between clients and the server.
    The message contains a payload, a type and a sender. The payload can be of
    any type. The type of a message is a string that describes the content of the message.
    The sender is the object that sends the message.

    Args:
        payload (Any): The content of the message.
        msg_type (str): The type of the message. Default is "model".
        sender (Optional[Any]): The sender of the message. Default is ``None``.
        inmemory (Optional[bool]): If ``True``, the payload is stored in memory. If ``False``, the
            payload is stored in the cache. If ``None``, the payload is stored according to the
            setting in the :class:`fluke.FlukeENV`. Note that when ``inmemory`` is ``False``,
            the payload is stored in the cache if and only if the environment is not in memory,
            otherwise the payload is stored in memory. Default is ``None``.

    Note:
        The payload can be stored in memory if and only if its type is :class:`torch.nn.Module`,
        :class:`torch.Tensor` or :class:`fluke.FlukeCache._ObjectRef`. In the latter case, the
        reference count of the object is incremented by one.

    Example:
        Let us consider a simple example where a client wants to send preparea  message to be sent
        to the server with the content "Hello". The client can create a message as follows:

        .. code-block:: python
            :linenos:

            message = Message(payload="Hello", msg_type="greeting", sender=client)
            # Then through a channel (of type Channel) the message can be sent to the server
            # channel.send(message, server)

    See Also:
        :class:`Channel`
    """

    def __init__(self,
                 payload: Any,
                 msg_type: str = "model",
                 sender: Optional[Any] = None,
                 inmemory: Optional[bool] = None):
        self.__id: str = str(uuid.uuid4().hex)
        self.__msg_type: str = msg_type
        self.__payload: Any = payload
        self.__sender: Optional[Any] = sender
        self.__size: int = self.__get_size(payload)
        self.__inmemory: bool = inmemory if inmemory is not None else True

        if not FlukeENV().is_inmemory() and not inmemory:
            if isinstance(payload, (torch.nn.Module, FlukeCache._ObjectRef, torch.Tensor)):
                self.__payload = cache_obj(payload, f"{self.__id}")
                self.__inmemory = False

    @property
    def id(self) -> int:
        """Get the unique identifier of the message.

        Returns:
            int: The unique identifier of the message.
        """
        return self.__id

    @property
    def msg_type(self) -> str:
        """Get the type of the message.

        Returns:
            str: The type of the message.
        """
        return self.__msg_type

    @property
    def payload(self) -> Any:
        """Get the payload of the message.

        Note:
            If the payload is stored in the cache, it is loaded in memory when this property is
            called for the first time on the object.

        Returns:
            Any: The payload of the message.
        """
        if not self.__inmemory:
            self.__payload = retrieve_obj(f"{self.__id}")
            self.__inmemory = True

        return self.__payload

    @property
    def sender(self) -> Optional[Any]:
        """Get the sender of the message.

        Returns:
            Optional[Any]: The sender of the message.
        """
        return self.__sender

    def __get_size(self, obj: Any) -> int:
        if obj is None or isinstance(obj, (int, float, bool)):
            return 1
        elif isinstance(obj, str):
            return len(obj)
        elif isinstance(obj, (list, tuple, set)):
            return sum([self.__get_size(i) for i in obj])
        elif isinstance(obj, dict):
            return self.__get_size(list(obj.values())) + self.__get_size(list(obj.keys()))
        elif isinstance(obj, np.ndarray):
            return obj.size
        elif isinstance(obj, torch.Tensor):
            return obj.numel()
        elif isinstance(obj, (torch.nn.Module)):
            return sum(p.numel() for p in obj.parameters())
        elif isinstance(obj, FlukeCache._ObjectRef):
            return 0
        else:
            warnings.warn(f"Unknown type {type(obj)} of object {obj} in payload." +
                          "Returning object size = 0.")
            return 0

    def clone(self, inmemory: Optional[bool] = None) -> Message:
        """Clone the message. The cloned message containes a deepcopy of the payload while
        keeping the same message type and the same reference to the sender.

        Args:
            inmemory (Optional[bool]): If ``True``, the payload is stored in memory. If False, the
                payload is stored in the cache. If None, the payload is stored according to the
                setting in the :class:`fluke.FlukeENV`.

        Returns:
            Message: The cloned message.
        """
        msg = Message(deepcopy(self.__payload), self.msg_type, self.sender, inmemory)
        msg.__size = self.__size
        return msg

    @property
    def size(self) -> int:
        """Get the size of the message. The message size is the size of the payload calculated in
        terms of "floating point" numbers. For example, a message containing a tensor of size
        (10, 10) has a size of 100. A message containing a string of length 10 has a size of 10.
        A message containing an ACK (i.e., with no payload) has a size of 1. In case of unknown
        types, a warning is raised and the size is set to 0.

        Returns:
            int: The size of the message in bytes.

        Example:

            .. code-block:: python
                :linenos:

                message = Message("Hello", "greeting", client)
                print(message.size)  # 5

                message = Message(torch.randn(10, 10), "tensor", client)
                print(message.size)  # 100

                message = Message(None, "ack", client)
                print(message.size)  # 1
        """
        return self.__size

    def __eq__(self, other: Message) -> bool:
        return self.__payload == other.__payload and self.msg_type == other.msg_type and \
            self.sender == other.sender

    def __str__(self) -> str:
        return f"Message[{self.id}](type={self.msg_type}, from={self.sender}, " + \
            f"payload={self.__payload}, size={self.size}, inmemory={self.__inmemory})"

    def __repr__(self) -> str:
        return str(self)

    def ram(self) -> bool:
        """Store the payload in memory.
        This method is used when the payload needs to be moved from the cache to memory.

        Returns:
            bool: ``True`` if the payload is moved to memory, ``False`` otherwise.
        """
        if not self.__inmemory:
            self.__payload = retrieve_obj(f"{self.__id}")
            self.__inmemory = True
            return True
        return False

    def cache(self) -> bool:
        """Store the payload in the cache.
        This method is used when the payload needs to be moved from memory to the cache.

        Returns:
            bool: ``True`` if the payload is moved to the cache, ``False`` otherwise.
        """
        if self.__inmemory:
            if isinstance(self.__payload, (torch.nn.Module, FlukeCache._ObjectRef, torch.Tensor)):
                self.__payload = cache_obj(self.__payload, f"{self.__id}")
                self.__inmemory = False
                return True
        return False


class ChannelObserver():
    """Channel observer interface for the Observer pattern.
    This interface is used to observe the communication channel during the federated learning
    process.

    See Also:
        :class:`fluke.utils.ServerObserver`, :class:`fluke.ObserverSubject`

    """

    def message_received(self, message: Message) -> None:
        """This method is called when a message is received, i.e., when a message is read from the
        message box of the receiver.

        Args:
            message (Message): The message received.
        """
        pass


class Channel(ObserverSubject):
    """A bi-directional communication channel. It is used to send and receive messages between the
    parties.
    The Channel class implements the Observer pattern. It notifies the observers when a message is
    received. Clients and server are supposed to use a channel to communicate with each other.

    Example:

        Let us consider a simple example where a ``client`` sends a message to the ``server``.

        .. code-block:: python
            :linenos:

            channel = Channel()
            channel.send(Message("Hello", "greeting", server), client)
            mag: Message = channel.receive(server, client, "greeting")
            print(msg.payload) # Hello
    """

    def __init__(self):
        super().__init__()
        self._buffer: dict[Any, list[Message]] = defaultdict(list)

    def __getitem__(self, mbox: Any) -> list[Message]:
        """Get the messages of the given receiver.

        Args:
            mbox (Any): The receiver.

        Returns:
            list[Message]: The list of messages sent to the receiver.
        """
        return self._buffer[mbox]

    @property
    def buffer(self) -> dict[Any, list[Message]]:
        """Get the buffer of the channel. The buffer stores the unread messages in a dictionary.
        The keys are the recipients and the values are the list of messages sent to the recipient.

        Returns:
            dict[Any, list[Message]]: The buffer of the channel.
        """
        return self._buffer

    def send(self, message: Message, mbox: Any) -> None:
        """Send a copy of the message to a receiver.
        To any sent message should correspond a received message. The receiver should call the
        `receive` method of the channel to get the message.

        Args:
            message (Message): The message to be sent.
            mbox (Any): The receiver.

        Example:
            Sending a string message from the ``server`` to a ``client``:

            .. code-block:: python
                :linenos:

                channel = Channel()
                channel.send(Message("Hello", "greeting", server), client)

        """
        self._buffer[mbox].append(message.clone())

    def receive(self, mbox: Any, sender: Any = None, msg_type: str = None) -> Message:
        """Receive (i.e., read) a message from a sender. The message is removed from the message box
        of the receiver. If both ``sender`` and ``msg_type`` are None, the first message in the
        message box is returned. If ``sender`` is None, the first message with the given
        ``msg_type`` is returned. If ``msg_type`` is None, the first message from the given
        ``sender`` is returned.

        Args:
            mbox (Any): The receiver.
            sender (Any): The sender.
            msg_type (str): The type of the message.

        Returns:
            Message: The received message.

        Raises:
            ValueError: message not found in the message box of the receiver with the
                given sender and message type.

        Example:
            Receiving a message from the ``server`` with message type ``greeting``:

            .. code-block:: python
                :linenos:

                channel = Channel()
                message = channel.receive(client, server, "greeting")
        """
        if sender is None and msg_type is None:
            msg = self._buffer[mbox].pop()
            self._notify_message_received(msg)
            return msg

        for i, msg in enumerate(self._buffer[mbox]):
            if sender is None or msg.sender == sender:  # match sender
                if msg_type is None or msg.msg_type == msg_type:  # match msg_type
                    msg = self._buffer[mbox].pop(i)
                    self._notify_message_received(msg)
                    return msg

        raise ValueError(f"Message from {sender} with msg type {msg_type} not found in {mbox}")

    def broadcast(self, message: Message, to: list[Any]) -> None:
        """Send a copy of the message to a list of receivers.

        Note:
            This method may have a side effect on the payload of the message.
            The payload is moved to memory after the message is sent to avoid dangling references.
            If the payload was already in memory, nothing is done.

        Important:
            When a message is broadcasted, a copy of the message is sent to each receiver. If the
            message is not stored in memory, the payload is a reference to an object stored in the
            cache, and thus the reference count of the object is incremented by the number of
            receivers. After the message is sent, the sent message is moved to memory to avoid
            dangling references.

        Args:
            message (Message): The message to be sent.
            to (list[Any]): The list of receivers.
        """
        for client in to:
            self.send(message, client)
        message.ram()

    def clear(self, mbox: Any) -> None:
        """Clear the message box of the given receiver.

        Caution:
            Any unread message will be lost after calling this method.
            Lost messages are not considered in the communication protocol thus they are not
            accounted for in the communication cost.

        Args:
            mbox (Any): The receiver.
        """
        if not FlukeENV().is_inmemory():
            cache = FlukeENV().get_cache()
            for msg in self._buffer[mbox]:
                cache.delete(f"{msg.id}")

        self._buffer[mbox].clear()

    def _notify_message_received(self, message: Message) -> None:
        """Notify the observers that a message has been received.

        Args:
            message (Message): The message received.
        """
        for observer in self._observers:
            observer.message_received(message)



"""Implementation of the [FedNova21]_ algorithm.

References:
    .. [FedNova21] Jianyu Wang, Qinghua Liu, Hao Liang, Gauri Joshi, and H. Vincent Poor.
       Tackling the Objective Inconsistency Problem in Heterogeneous Federated Optimization.
       In NeurIPS 2020. URL: https://arxiv.org/abs/2007.07481
"""
import sys
from typing import Any, Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from ..utils.model import aggregate_models  # NOQA

__all__ = [
    "FedNovaClient",
    "FedNovaServer",
    "FedNova"
]


class FedNovaClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.tau = 0

    def _get_momentum(self):
        if self.optimizer is None:
            if "momentum" in self._optimizer_cfg.optimizer_kwargs:
                return self._optimizer_cfg.optimizer_kwargs["momentum"]
            else:
                return 0
        else:
            return self.optimizer.param_groups[0]["momentum"]

    def fit(self, override_local_epochs: int = 0) -> float:
        loss = super().fit(override_local_epochs)
        self.tau = self.hyper_params.local_epochs * self.train_set.n_batches
        rho = self._get_momentum()
        self.a = (self.tau - rho * (1.0 - pow(rho, self.tau)) / (1.0 - rho)) / (1.0 - rho)
        self.channel.send(Message(self.a, "local_a", self, inmemory=True), self.server)
        return loss


class FedNovaServer(Server):

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        weights = self._get_client_weights(eligible)
        a_i = [
            self.channel.receive(self, client, "local_a").payload
            for client in eligible
        ]
        coeff = sum([a_i[i] * weights[i] for i in range(len(eligible))])
        weights = torch.true_divide(torch.tensor(weights) * coeff, torch.tensor(a_i))
        aggregate_models(self.model, client_models, weights, self.hyper_params.lr)


class FedNova(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedNovaClient

    def get_server_class(self) -> Server:
        return FedNovaServer



"""`fluke` command line interface."""

from rich.console import Console
import os
import sys
import uuid
from typing import Any, List

import numpy as np
import pandas as pd
import rich
import torch
import typer
import yaml
from rich.panel import Panel
from rich.pretty import Pretty
from rich.progress import track

sys.path.append(".")

from . import FlukeENV  # NOQA
from .data import DataSplitter, FastDataLoader  # NOQA
from .data.datasets import Datasets  # NOQA
from .evaluation import ClassificationEval  # NOQA
from .utils import (Configuration, OptimizerConfigurator,  # NOQA
                    get_class_from_qualified_name, get_loss, get_model)
from .utils.log import get_logger  # NOQA

__version__ = "0.7.1"

console = Console()


def version_callback(value: bool):
    if value:
        print(f"fluke: {__version__}")
        raise typer.Exit()


app = typer.Typer()


@app.command()
def centralized(exp_cfg: str = typer.Argument(..., help="Configuration file"),
                alg_cfg: str = typer.Argument(..., help='Config file for the algorithm to run'),
                epochs: int = typer.Option(0, help='Number of epochs to run')) -> None:
    """Run a centralized learning experiment."""

    cfg = Configuration(exp_cfg, alg_cfg)
    FlukeENV().configure(cfg)
    data_container = Datasets.get(**cfg.data.dataset)

    device = FlukeENV().get_device()
    train_loader = FastDataLoader(*data_container.train,
                                  batch_size=cfg.client.batch_size,
                                  num_labels=data_container.num_classes,
                                  shuffle=True)
    test_loader = FastDataLoader(*data_container.test,
                                 batch_size=10,
                                 num_labels=data_container.num_classes,
                                 shuffle=False)

    hp = cfg.method.hyperparameters
    model = get_model(mname=hp.model, **hp.net_args if "net_args" in hp else {})
    if "name" not in hp.client.optimizer:
        hp.client.optimizer.name = torch.optim.SGD
    optimizer_cfg = OptimizerConfigurator(optimizer_cfg=hp.client.optimizer,
                                          scheduler_cfg=hp.client.scheduler)
    optimizer, scheduler = optimizer_cfg(model)
    criterion = get_loss(hp.client.loss)
    evaluator = ClassificationEval(eval_every=cfg.eval.eval_every,
                                   n_classes=data_container.num_classes)
    history = []

    model.to(device)
    epochs = epochs if epochs > 0 else int(
        max(1, cfg.protocol.n_rounds * cfg.protocol.eligible_perc))

    exp_name = f"Centralized_{cfg.data.dataset.name}_E{epochs}_S{cfg.exp.seed}"
    log = get_logger(cfg.logger.name, name=exp_name, **cfg.logger.exclude('name'))
    log.init(**cfg, exp_id=uuid.uuid4().hex)
    log.log(f"Centralized Learning [ #Epochs = {epochs} ]\n")

    for e in range(epochs):
        model.train()
        rich.print(f"Epoch {e+1}")
        for _, (X, y) in track(enumerate(train_loader),
                               total=train_loader.n_batches,
                               transient=True):
            X, y = X.to(device), y.to(device)
            optimizer.zero_grad()
            y_hat = model(X)
            loss = criterion(y_hat, y)
            loss.backward()
            optimizer.step()
        scheduler.step()

        epoch_eval = evaluator.evaluate(e+1, model, test_loader, criterion, device=device)
        history.append(epoch_eval)
        for k, v in epoch_eval.items():
            log.add_scalar(k, v, e+1)
        log.pretty_log(epoch_eval, title=f"Performance [Epoch {e+1}]")
        rich.print()
    model.cpu()


@app.command()
def federation(exp_cfg: str = typer.Argument(..., help="Configuration file"),
               alg_cfg: str = typer.Argument(..., help='Config file for the algorithm to run'),
               resume: str = typer.Option(None,
                                          help='Path to the checkpoint file to load.')) -> None:
    """Run a federated learning experiment."""

    cfg = Configuration(exp_cfg, alg_cfg)
    _run_federation(cfg, resume)


@app.command()
def sweep(exp_cfg: str = typer.Argument(..., help="Configuration file"),
          alg_cfgs: List[str] = typer.Argument(...,
                                               help='Config file(s) for the algorithm(s) to run')
          ) -> None:
    """Run a battery of federated learning experiments with different configurations."""

    for alg_cfg in alg_cfgs:
        for cfg in Configuration.sweep(exp_cfg, alg_cfg):
            try:
                _run_federation(cfg)
            except Exception as e:
                # print the error with rich
                rich.print(f"[yellow]Error with experiment: {cfg.verbose()}")
                rich.print(f"[red]{e}")
                console.print_exception(max_frames=10)
                continue


def _run_federation(cfg: Configuration, resume: str = None) -> None:
    FlukeENV().configure(cfg)
    data_container = Datasets.get(**cfg.data.dataset)
    evaluator = ClassificationEval(eval_every=cfg.eval.eval_every,
                                   n_classes=data_container.num_classes)
    FlukeENV().set_evaluator(evaluator)

    data_splitter = DataSplitter(dataset=data_container,
                                 distribution=cfg.data.distribution.name,
                                 dist_args=cfg.data.distribution.exclude("name"),
                                 **cfg.data.exclude('dataset', 'distribution'))

    fl_algo_class = get_class_from_qualified_name(cfg.method.name)
    fl_algo = fl_algo_class(cfg.protocol.n_clients,
                            data_splitter,
                            cfg.method.hyperparameters)
    # plot_distribution(fl_algo.clients)
    if cfg.save and cfg.save.path:
        path = f"{cfg.save.path}_{fl_algo.id}"
        if not os.path.exists(path):
            os.makedirs(path)
        yaml.dump(cfg.to_dict(), open(f"{path}/config.yaml", "w"))

    log_name = f"{fl_algo.__class__.__name__} [{fl_algo.id}]"
    log = get_logger(cfg.logger.name, name=log_name, **cfg.logger.exclude('name'))
    log.init(**cfg, exp_id=fl_algo.id)

    fl_algo.set_callbacks([log])
    rich.print(Panel(Pretty(fl_algo), title="FL algorithm"))

    if resume is not None:
        fl_algo.load(resume)

    try:
        fl_algo.run(cfg.protocol.n_rounds, cfg.protocol.eligible_perc)
    except Exception as e:
        log.log(f"Error: {e}")
        FlukeENV().force_close()
        FlukeENV.clear()
        log.close()
        FlukeENV().close_cache()
        raise e

    FlukeENV().close_cache()
    log.close()


@app.command()
def clients_only(exp_cfg: str = typer.Argument(..., help="Configuration file"),
                 alg_cfg: str = typer.Argument(..., help='Config file for the algorithm to run'),
                 epochs: int = typer.Option(0, help='Number of epochs to run')) -> None:
    """Run a local training (for all clients) experiment."""

    cfg = Configuration(exp_cfg, alg_cfg)
    FlukeENV().configure(cfg)

    data_container = Datasets.get(**cfg.data.dataset)
    data_splitter = DataSplitter(dataset=data_container,
                                 distribution=cfg.data.distribution.name,
                                 dist_args=cfg.data.distribution.exclude("name"),
                                 **cfg.data.exclude('dataset', 'distribution'))

    device = FlukeENV().get_device()

    hp = cfg.method.hyperparameters
    (clients_tr_data, clients_te_data), _ = \
        data_splitter.assign(cfg.protocol.n_clients, hp.client.batch_size)

    criterion = get_loss(hp.client.loss)
    client_evals = []
    if epochs == 0:
        epochs = max(100, int(cfg.protocol.n_rounds *
                              hp.client.local_epochs * cfg.protocol.eligible_perc))
    progress = track(enumerate(zip(clients_tr_data, clients_te_data)),
                     total=len(clients_tr_data),
                     description="Clients training...")
    exp_name = "Clients-only_" + "_".join(str(cfg).split("_")[1:])
    log = get_logger(cfg.logger.name, name=exp_name, **cfg.logger.exclude('name'))
    log.init(**cfg, exp_id=uuid.uuid4().hex)

    running_evals = {c: [] for c in range(cfg.protocol.n_clients)}
    for i, (train_loader, test_loader) in progress:
        log.log(f"Client [{i+1}/{cfg.protocol.n_clients}]")
        model = get_model(mname=hp.model, **hp.net_args if "net_args" in hp else {})
        hp.client.optimizer.name = torch.optim.SGD
        optimizer_cfg = OptimizerConfigurator(optimizer_cfg=hp.client.optimizer,
                                              scheduler_cfg=hp.client.scheduler)
        optimizer, scheduler = optimizer_cfg(model)
        evaluator = ClassificationEval(eval_every=cfg.eval.eval_every,
                                       n_classes=data_container.num_classes)
        model.to(device)
        for e in range(epochs):
            model.train()
            for _, (X, y) in enumerate(train_loader):
                X, y = X.to(device), y.to(device)
                optimizer.zero_grad()
                y_hat = model(X)
                loss = criterion(y_hat, y)
                loss.backward()
                optimizer.step()
            scheduler.step()

            client_eval = evaluator.evaluate(e+1, model, test_loader, criterion, device=device)
            running_evals[i].append(client_eval)

        log.pretty_log(client_eval, title=f"Client [{i}] Performance")
        client_evals.append(client_eval)
        model.cpu()

    for e in range(epochs):
        for c in running_evals:
            log.add_scalars(f"Client[{c}]", running_evals[c][e], e+1)

    client_mean = pd.DataFrame(client_evals).mean(numeric_only=True).to_dict()
    client_mean = {k: float(np.round(float(v), 5)) for k, v in client_mean.items()}
    log.pretty_log(client_mean, title="Overall local performance")


@app.callback()
def run(version: bool = typer.Option(None, "--version", help="Show the installed version of fluke",
                                     callback=version_callback)) -> None:
    pass


def main() -> Any:
    return app()


if __name__ == "__main__":
    main()



"""
The :mod:`fluke` module is the entry module of the :mod:`fluke` framework. Here are defined generic
classes used by the other modules.
"""
from __future__ import annotations

import random
import re
import shutil
import uuid
import warnings
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Iterable, Union

import numpy as np
import torch
from diskcache import Cache
from rich.console import Group
from rich.progress import Live, Progress

if TYPE_CHECKING:
    from .evaluation import Evaluator


__all__ = [
    'algorithms',
    'client',
    'comm',
    'data',
    'evaluation',
    'get',
    'nets',
    'run',
    'server',
    'utils',
    'DDict',
    'FlukeCache',
    'FlukeENV',
    'ObserverSubject',
    'Singleton'
]


class Singleton(type):
    """This metaclass is used to create singleton classes. A singleton class is a class that can
    have only one instance. If the instance does not exist, it is created; otherwise, the existing
    instance is returned.

    Example:
        .. code-block:: python
            :linenos:

            class MyClass(metaclass=Singleton):
                pass
            a = MyClass()
            b = MyClass()
            print(a is b)  # True

    """
    _instances = {}

    def __call__(cls, *args, **kwargs: dict[str, Any]):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    def clear(cls):
        cls._instances = {}


class DDict(dict):
    """A dictionary that can be accessed with dot notation recursively.

    Important:
        The :class:`DDict` is a subclass of the built-in :class:`dict` class and it behaves like a
        dictionary. However, the keys must be strings.

    Example:
        .. code-block:: python
            :linenos:

            d = DDict(a=1, b=2, c={'d': 3, 'e': 4})
            print(d.a)  # 1
            print(d.b)  # 2
            print(d.c.d)  # 3
            print(d.c.e)  # 4

    """
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

    def __init__(self, *args: dict, **kwargs: dict[str, Any]):
        self.update(*args, **kwargs)

    def update(self, *args: dict, **kwargs: dict[str, Any]):
        """Update the :class:`DDict` with the specified key-value pairs.

        Args:
            *args (dict): Dictionary with the key-value pairs.
            **kwargs: The key-value pairs.

        Example:
            .. code-block:: python
                :linenos:

                d = DDict(a=1)
                print(d) # {'a': 1}
                d.update(b=2, c=3)
                print(d) # {'a': 1, 'b': 2, 'c': 3}

        """
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    if isinstance(v, dict):
                        self[k] = DDict(**v)
                    else:
                        self[k] = v
            else:
                warnings.warn(f"Argument {arg} is not a dictionary and will be ignored.")

        for k, v in kwargs.items():
            if isinstance(v, dict):
                self[k] = DDict(**v)
            else:
                self[k] = v

    def exclude(self, *keys: str):
        """Create a new :class:`DDict` excluding the specified keys.

        Args:
            *keys: The keys to be excluded.

        Returns:
            DDict: The new DDict.

        Example:
            .. code-block:: python
                :linenos:

                d = DDict(a=1, b=2, c=3)
                e = d.exclude('b', 'c')
                print(e) # {'a': 1}

        """
        return DDict(**{k: v for k, v in self.items() if k not in keys})

    def match(self, other: DDict, full: bool = True) -> bool:
        """Check if the two :class:`DDict` match.

        Args:
            other (DDict): The other :class:`DDict`.
            full (bool): If ``True``, the two :class:`DDict` must match exactly. If ``False``, the
                `other` :class:`DDict` must be a subset of the current :class:`DDict`.

        Returns:
            bool: Whether the two :class:`DDict` match.
        """
        if full:
            return self == other
        return all(k in self and (self[k] == other[k] if not isinstance(self[k], DDict)
                                  else self[k].match(other[k], False)) for k in other.keys())

    def diff(self, other: DDict) -> DDict:
        """Get the difference between two :class:`DDict`.

        Args:
            other (DDict): The other :class:`DDict`.

        Returns:
            DDict: The difference between the two :class:`DDict`.

        Example:
            .. code-block:: python
                :linenos:

                d = DDict(a=1, b=2, c=3)
                e = DDict(a=1, b=3, c=4)
                print(d.diff(e)) # {'b': 3, 'c': 4}
        """
        diff = DDict()
        for k, v in other.items():
            if k in self:
                if isinstance(self[k], DDict):
                    d = self[k].diff(v)
                    if d:
                        diff[k] = d
                elif v != self[k]:
                    diff[k] = v
            else:
                diff[k] = v
        return diff

    def __getstate__(self) -> dict:
        return self.__dict__

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)


class ObserverSubject():
    """Subject class for the observer pattern. The subject is the class that is observed and thus
    it holds the observers.

    Example:
        .. code-block:: python
            :linenos:

            class MySubject(ObserverSubject):
                def __init__(self):
                    super().__init__()
                    self._data = 0

                @property
                def data(self):
                    return self._data

                @data.setter
                def data(self, value):
                    self._data = value
                    self.notify()

            class MyObserver:
                def __init__(self, subject):
                    subject.attach(self)

                def update(self):
                    print("Data changed.")

            subject = MySubject()
            observer = MyObserver(subject)
            subject.data = 1  # "Data changed."

    """

    def __init__(self):
        self._observers: list[Any] = []

    def attach(self, observer: Union[Any, Iterable[Any]]):
        """Attach one or more observers.

        Args:
            observer (Union[Any, Iterable[Any]]): The observer or a list of observers.
        """
        if observer is None:
            return

        if not isinstance(observer, (list, tuple, set)):
            observer = [observer]

        for obs in observer:
            if obs not in self._observers:
                self._observers.append(obs)

    def detach(self, observer: Any):
        """Detach an observer.

        Args:
            observer (Any): The observer to be detached.
        """
        try:
            self._observers.remove(observer)
        except ValueError:
            pass


class FlukeENV(metaclass=Singleton):
    """Environment class for the :mod:`fluke` framework.
    This class is a singleton and it contains environment settings that are used by the other
    classes. The environment includes:

    - The device (``"cpu"``, ``"cuda[:N]"``, ``"auto"``, ``"mps"``);
    - The ``seed`` for reproducibility;
    - If the models are stored in memory or on disk (when not in use);
    - The evaluation configuration;
    - The saving settings;
    - The progress bars for the federated learning process, clients and the server;
    - The live renderer, which is used to render the progress bars.

    """

    # general settings
    _device: torch.device = torch.device('cpu')
    _seed: int = 0
    _inmemory: bool = True
    _cache: FlukeCache = None

    # saving settings
    _save_path: str = None
    _save_every: int = 0
    _global_only: bool = False
    _temp_path: str = None

    # evaluation settings
    _evaluator: Evaluator = None
    _eval_cfg: dict = {
        "pre_fit": False,
        "post_fit": False,
        "locals": False,
        "server": True
    }

    # progress bars
    _rich_progress_FL: Progress = None
    _rich_progress_clients: Progress = None
    _rich_progress_server: Progress = None
    _live_renderer: Live = None

    def __init__(self):
        super().__init__()
        self._rich_progress_FL: Progress = Progress(transient=True)
        self._rich_progress_clients: Progress = Progress(transient=True)
        self._rich_progress_server: Progress = Progress(transient=True)
        self._rich_live_renderer: Live = Live(Group(self._rich_progress_FL,
                                                    self._rich_progress_clients,
                                                    self._rich_progress_server))

    def configure(self, cfg: DDict) -> None:
        """Configure the global settings.

        Args:
            config (DDict): The configuration.
        """
        self.set_seed(cfg.exp.seed)
        self.set_device(cfg.exp.device)
        self.set_inmemory(cfg.exp.inmemory)
        self.set_save_options(**cfg.save)
        self.set_eval_cfg(cfg.eval)

    def get_seed(self) -> int:
        """Get the seed.

        Returns:
            int: The seed.
        """
        return self._seed

    def get_eval_cfg(self) -> DDict:
        """Get the evaluation configuration.

        Returns:
            DDict: The evaluation configuration.
        """
        return DDict(self._eval_cfg)

    def set_eval_cfg(self, cfg: DDict) -> None:
        """Set the evaluation configuration.

        Args:
            cfg (DDict): The evaluation configuration.
        """
        for key, value in cfg.items():
            self._eval_cfg[key] = value

    def get_evaluator(self) -> Evaluator:
        """Get the evaluator.

        Returns:
            Evaluator: The evaluator.
        """
        return self._evaluator

    def set_evaluator(self, evaluator: Evaluator) -> None:
        """Set the evaluator.

        Args:
            evaluator (Evaluator): The evaluator.
        """
        self._evaluator = evaluator

    def set_seed(self, seed: int) -> None:
        """Set seed for reproducibility. The seed is used to set the random seed for the following
        libraries: ``torch``, ``torch.cuda``, ``numpy``, ``random``.

        Args:
            seed (int): The seed.
        """
        self._seed = seed
        torch.manual_seed(seed)
        gen = torch.Generator()
        gen.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)
        random.seed(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

    def auto_device(self) -> torch.device:
        """Set device to ``cuda`` or ``mps`` if available, otherwise ``cpu``.

        Returns:
            torch.device: The device.
        """
        if torch.cuda.is_available():
            self._device = torch.device('cuda')
        elif torch.backends.mps.is_available():
            self._device = torch.device('mps')
        else:
            self._device = torch.device('cpu')
        return self._device

    def set_device(self, device: str) -> torch.device:
        """Set the device. The device can be ``cpu``, ``auto``, ``mps``, ``cuda`` or ``cuda:N``,
        where ``N`` is the GPU index.

        Args:
            device (str): The device as string.

        Returns:
            torch.device: The selected device as torch.device.
        """
        assert device in ['cpu', 'auto', 'mps', 'cuda'] or re.match(r'^cuda:\d+$', device), \
            f"Invalid device {device}."

        if device == "auto":
            return FlukeENV().auto_device()

        if device.startswith('cuda') and ":" in device:
            idx = int(device.split(":")[1])
            self._device = torch.device("cuda", idx)
        else:
            self._device = torch.device(device)
        return self._device

    def get_device(self) -> torch.device:
        """Get the current device.

        Returns:
            torch.device: The device.
        """
        return self._device

    def get_progress_bar(self, progress_type: str) -> Progress:
        """Get the progress bar.
        The possible progress bar types are:

        - ``FL``: The progress bar for the federated learning process.
        - ``clients``: The progress bar for the clients.
        - ``server``: The progress bar for the server.

        Args:
            progress_type (str): The type of progress bar.

        Returns:
            Progress: The progress bar.

        Raises:
            ValueError: If the progress bar type is invalid.
        """
        if progress_type == 'FL':
            return self._rich_progress_FL
        elif progress_type == 'clients':
            return self._rich_progress_clients
        elif progress_type == 'server':
            return self._rich_progress_server
        else:
            raise ValueError(f'Invalid type of progress bar type {progress_type}.')

    def get_live_renderer(self) -> Live:
        """Get the live renderer. The live renderer is used to render the progress bars.

        Returns:
            Live: The live renderer.
        """
        return self._rich_live_renderer

    def get_save_options(self) -> tuple[str, int, bool]:
        """Get the save options.

        Returns:
            tuple[str, int, bool]: The save path, the save frequency and the global only flag.
        """
        return self._save_path, self._save_every, self._global_only

    def set_save_options(self,
                         path: str | None = None,
                         save_every: int | None = None,
                         global_only: bool | None = None) -> None:
        """Set the save options.

        Args:
            path (str): The path to save the checkpoints.
            save_every (int): The frequency of saving the checkpoints.
            global_only (bool): If ``True``, only the global model is saved.
        """
        if path is not None:
            self._save_path = path
        if save_every is not None:
            self._save_every = save_every
        if global_only is not None:
            self._global_only = global_only

    def set_inmemory(self, inmemory: bool) -> None:
        """Set if the data is stored in memory.

        Args:
            inmemory (bool): If ``True``, the data is stored in memory, otherwise it is stored on
                disk.
        """
        self._inmemory = inmemory

    def get_cache(self) -> FlukeCache:
        """Get the cache.

        Returns:
            Cache: The cache.
        """
        return self._cache

    def open_cache(self, path: str) -> None:
        """Open the cache at the specified path if the ``inmemory`` flag is ``False``.

        Note:
            The full path to the cache is ``tmp/path`` where ``path`` is the specified path.
            We suggest to use as path the UUID of the experiment.

        Args:
            path (str): The path to the cache.
        """
        if not self._inmemory and self._cache is None:
            self._cache = FlukeCache(path)
        elif self._cache is not None:
            warnings.warn("Cache already open.")

    def close_cache(self) -> None:
        """Close the cache."""
        if self._cache is not None:
            self._cache.close()
            self._cache = None

    def is_inmemory(self) -> bool:
        """Check if the data is stored in memory.

        Returns:
            bool: If ``True``, the data is stored in memory, otherwise it is stored on disk.
        """
        return self._inmemory

    def force_close(self) -> None:
        """Force close the progress bars and the live renderer."""

        task_ids = [task.id for task in self._rich_progress_FL.tasks]
        for tid in task_ids:
            self._rich_progress_FL.remove_task(tid)

        task_ids = [task.id for task in self._rich_progress_clients.tasks]
        for tid in task_ids:
            self._rich_progress_clients.remove_task(tid)

        task_ids = [task.id for task in self._rich_progress_server.tasks]
        for tid in task_ids:
            self._rich_progress_server.remove_task(tid)

        self._rich_live_renderer.refresh()
        self._rich_live_renderer.stop()

    def __getstate__(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if not k.startswith('_rich')}

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)


class FlukeCache():
    """A cache class that can store data on disk."""

    class _ObjectRef():
        """A reference to an object in the cache.
        The reference is a unique identifier that is used to store and retrieve the object from the
        cache.
        """

        def __init__(self):
            self._id = str(uuid.uuid4().hex)

        @property
        def id(self) -> str:
            """Get the unique identifier of the reference.

            Returns:
                str: The unique identifier.
            """
            return self._id

    class _RefCounter():
        """A reference counter for an object in the cache."""

        def __init__(self, value: Any, refs: int = 1):
            self._value = value
            self._refs = refs
            self._id = FlukeCache._ObjectRef()

        @property
        def id(self) -> FlukeCache._ObjectRef:
            """Get the unique identifier of the reference.

            Returns:
                str: The unique identifier.
            """
            return self._id

        @property
        def value(self) -> Any:
            """Get the value ppinted by the reference.

            Returns:
                Any: The value.
            """
            return self._value

        @property
        def refs(self) -> int:
            """Get the number of references to the object in the cache.

            Returns:
                int: The number of references.
            """
            return self._refs

        def dec(self) -> Any:
            """Decrement the number of references to the object in the cache.

            Returns:
                FlukeCache._RefCounter: The reference counter.
            """
            self._refs -= 1
            return self

        def inc(self) -> FlukeCache._RefCounter:
            """Increment the number of references to the object in the cache.

            Returns:
                FlukeCache._RefCounter: The reference counter.
            """
            self._refs += 1
            return self

    def __init__(self, path: str, **kwargs):
        if 'size_limit' not in kwargs:
            kwargs['size_limit'] = 2**34
        self._cache: Cache = Cache(f"tmp/{path}", **kwargs)
        self._key2ref: dict[str, FlukeCache._ObjectRef] = {}

    def __getitem__(self, key: str):
        return self._cache[self._key2ref[key].id].value

    @property
    def cache_dir(self) -> str:
        """Get the cache directory.

        Returns:
            str: The cache directory.
        """
        return self._cache.directory

    def get(self, key: str, default: Any = None):
        """Get the object identified by the key from the cache.

        If the object is not in the cache, the default value is returned.

        Note:
            The object is still in the cache after this operation.

        Args:
            key (str): The key of the object.
            default (Any, optional): The default value to return if the object is not in the cache.
                Defaults to None.

        Returns:
            Any: The object in the cache or the default value.
        """
        if key not in self._key2ref:
            return default
        obj = self._cache.get(self._key2ref[key].id, default=default)
        if obj is not default:
            return obj.value

    def push(self, key: str, value: Any) -> FlukeCache._ObjectRef:
        """Push an object to the cache.

        Note:
            If the object that is pushed is already a cache reference, then the referenced object is
            already in the cache and its reference counter is incremented.

        Args:
            key (str): The key of the object.
            value (Any): The object to store in the cache.

        Returns:
            FlukeCache._ObjectRef: The reference to the object in the cache.
        """
        if isinstance(value, FlukeCache._ObjectRef):
            assert value.id in self._cache, f"Reference {value.id} not in cache."
            self._key2ref[key] = value
            self._cache[value.id] = self._cache[value.id].inc()
            return value
        else:
            ref = self._RefCounter(value)
            self._key2ref[key] = ref.id
            self._cache[ref.id.id] = ref
            return ref.id

    def pop(self, key: str, copy: bool = True) -> Any:
        """Pop an object from the cache given its key.

        If the key is not in the cache, ``None`` is returned.

        Args:
            key (str): The key of the object.
            copy (bool, optional): If ``True``, a copy of the object is returned.
                Defaults to ``True``.

        Returns:
            Any: The object in the cache or its copy.
        """
        if key not in self._key2ref:
            return None
        ref = self._key2ref[key]
        del self._key2ref[key]
        self._cache[ref.id] = self._cache[ref.id].dec()
        obj = self._cache[ref.id].value
        if self._cache[ref.id].refs == 0:
            self._cache.delete(ref.id)
        return obj if not copy else deepcopy(obj)

    def delete(self, key: str) -> None:
        """Remove an object from the cache without returning it.

        If the key is not in the cache, nothing happens.

        Args:
            key (str): The key of the object.
        """
        if key in self._key2ref:
            ref = self._key2ref[key]
            del self._key2ref[key]
            self._cache[ref.id] = self._cache[ref.id].dec()
            if self._cache[ref.id].refs == 0:
                self._cache.delete(ref.id)

    def close(self) -> None:
        """Close the cache."""
        if self._cache is not None:
            self._cache.clear()
            self._cache.close()
            try:
                shutil.rmtree(self._cache.directory)
            except OSError:  # Windows wonkiness
                pass
            self._cache = None
            self._key2ref = {}

    @property
    def occupied(self) -> int:
        """Get the number of different objects in the cache.

        Returns:
            int: The number of objects in the cache.
        """
        return len(list(self._cache.iterkeys()))

    def cleanup(self) -> None:
        """Clean up the cache by removing the objects that are not referenced.

        This operation should not be necessary if the cache is used correctly.
        """
        keys = set([v.id for v in self._key2ref.values()])
        for key in self._cache.iterkeys():
            if key not in keys:
                self._cache.pop(key)



"""This module contains the definition of the evaluation classes used to perform the evaluation
of the model client-side and server-side."""
import sys
from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional, Union

import numpy as np
import torch
from torch.nn import Module
from torchmetrics import Accuracy, F1Score, Precision, Recall

sys.path.append(".")
sys.path.append("..")

from .data import FastDataLoader  # NOQA

__all__ = [
    "Evaluator",
    "ClassificationEval"
]


class Evaluator(ABC):
    """This class is the base class for all evaluators in :mod:`fluke`.
    An evaluator object should be used to perform the evaluation of a (federated) model.

    Args:
        eval_every (int): The evaluation frequency expressed as the number of rounds between
          two evaluations. Defaults to 1, i.e., evaluate the model at each round.

    Attributes:
        eval_every (int): The evaluation frequency.
    """

    def __init__(self, eval_every: int = 1):
        self.eval_every: int = eval_every

    @abstractmethod
    def evaluate(self,
                 round: int,
                 model: Module,
                 eval_data_loader: FastDataLoader,
                 loss_fn: Optional[torch.nn.Module],
                 **kwargs: dict[str, Any]) -> dict[str, Any]:
        """Evaluate the model.

        Args:
            round (int): The current
            model (Module): The model to evaluate.
            eval_data_loader (FastDataLoader): The data loader to use for evaluation.
            loss_fn (torch.nn.Module, optional): The loss function to use for evaluation.
            **kwargs: Additional keyword arguments.

        Returns:
            dict[str, Any]: A dictionary containing the computed metrics.
        """
        raise NotImplementedError

    def __call__(self,
                 round: int,
                 model: Module,
                 eval_data_loader: FastDataLoader,
                 loss_fn: Optional[torch.nn.Module],
                 **kwargs: dict[str, Any]) -> dict:
        """Evaluate the model.

        Note:
            This method is equivalent to ``evaluate``.

        Args:
            round (int): The current round.
            model (Module): The model to evaluate.
            eval_data_loader (FastDataLoader): The data loader to use for evaluation.
            loss_fn (torch.nn.Module, optional): The loss function to use for evaluation.
            **kwargs: Additional keyword arguments.

        Returns:
            dict[str, Any]: A dictionary containing the computed metrics.
        """
        return self.evaluate(round=round,
                             model=model,
                             eval_data_loader=eval_data_loader,
                             loss_fn=loss_fn,
                             **kwargs)


class ClassificationEval(Evaluator):
    """Evaluate a PyTorch model for classification.
    The metrics computed are ``accuracy``, ``precision``, ``recall``, ``f1`` and the loss according
    to the provided loss function ``loss_fn`` when calling the method ``evaluation``.
    Metrics are computed both in a micro and macro fashion.

    Args:
        eval_every (int): The evaluation frequency.
        n_classes (int): The number of classes.

    Attributes:
        eval_every (int): The evaluation frequency.
        n_classes (int): The number of classes.
    """

    def __init__(self, eval_every: int, n_classes: int):
        super().__init__(eval_every=eval_every)
        self.n_classes: int = n_classes

    @torch.no_grad
    def evaluate(self,
                 round: int,
                 model: torch.nn.Module,
                 eval_data_loader: Union[FastDataLoader,
                                         Iterable[FastDataLoader]],
                 loss_fn: Optional[torch.nn.Module] = None,
                 device: torch.device = torch.device("cpu")) -> dict:
        """Evaluate the model. The metrics computed are ``accuracy``, ``precision``, ``recall``,
        ``f1`` and the loss according to the provided loss function ``loss_fn``. Metrics are
        computed both in a micro and macro fashion.

        Warning:
            The loss function ``loss_fn`` should be defined on the same device as the model.
            Moreover, it is assumed that the only arguments of the loss function are the predicted
            values and the true values.

        Args:
            round (int): The current round.
            model (torch.nn.Module): The model to evaluate. If ``None``, the method returns an
                empty dictionary.
            eval_data_loader (Union[FastDataLoader, Iterable[FastDataLoader]]):
                The data loader(s) to use for evaluation. If ``None``, the method returns an empty
                dictionary.
            loss_fn (torch.nn.Module, optional): The loss function to use for evaluation.
            device (torch.device, optional): The device to use for evaluation. Defaults to "cpu".

        Returns:
            dict: A dictionary containing the computed metrics.
        """
        from .utils import clear_cuda_cache  # NOQA

        if (round != 1) and (round % self.eval_every != 0):
            return {}

        if (model is None) or (eval_data_loader is None):
            return {}

        model.eval()
        model.to(device)
        task = "multiclass"  # if self.n_classes >= 2 else "binary"
        accs, losses = [], []
        micro_precs, micro_recs, micro_f1s = [], [], []
        macro_precs, macro_recs, macro_f1s = [], [], []
        loss, cnt = 0, 0

        if not isinstance(eval_data_loader, list):
            eval_data_loader = [eval_data_loader]

        for data_loader in eval_data_loader:
            accuracy = Accuracy(task=task, num_classes=self.n_classes, top_k=1, average="micro")
            micro_precision = Precision(
                task=task, num_classes=self.n_classes, top_k=1, average="micro")
            micro_recall = Recall(task=task, num_classes=self.n_classes, top_k=1, average="micro")
            micro_f1 = F1Score(task=task, num_classes=self.n_classes, top_k=1, average="micro")
            macro_precision = Precision(
                task=task, num_classes=self.n_classes, top_k=1, average="macro")
            macro_recall = Recall(task=task, num_classes=self.n_classes, top_k=1, average="macro")
            macro_f1 = F1Score(task=task, num_classes=self.n_classes, top_k=1, average="macro")
            loss = 0
            for X, y in data_loader:
                X, y = X.to(device), y.to(device)
                with torch.no_grad():
                    y_hat = model(X)
                    if loss_fn is not None:
                        loss += loss_fn(y_hat, y).item()

                accuracy.update(y_hat.cpu(), y.cpu())
                micro_precision.update(y_hat.cpu(), y.cpu())
                micro_recall.update(y_hat.cpu(), y.cpu())
                micro_f1.update(y_hat.cpu(), y.cpu())
                macro_precision.update(y_hat.cpu(), y.cpu())
                macro_recall.update(y_hat.cpu(), y.cpu())
                macro_f1.update(y_hat.cpu(), y.cpu())

            cnt += len(data_loader)
            accs.append(accuracy.compute().item())
            micro_precs.append(micro_precision.compute().item())
            micro_recs.append(micro_recall.compute().item())
            micro_f1s.append(micro_f1.compute().item())
            macro_precs.append(macro_precision.compute().item())
            macro_recs.append(macro_recall.compute().item())
            macro_f1s.append(macro_f1.compute().item())
            losses.append(loss / cnt)

        model.cpu()
        clear_cuda_cache()

        result = {
            "accuracy":  np.round(sum(accs) / len(accs), 5).item(),
            "micro_precision": np.round(sum(micro_precs) / len(micro_precs), 5).item(),
            "micro_recall":    np.round(sum(micro_recs) / len(micro_recs), 5).item(),
            "micro_f1":        np.round(sum(micro_f1s) / len(micro_f1s), 5).item(),
            "macro_precision": np.round(sum(macro_precs) / len(macro_precs), 5).item(),
            "macro_recall":    np.round(sum(macro_recs) / len(macro_recs), 5).item(),
            "macro_f1":        np.round(sum(macro_f1s) / len(macro_f1s), 5).item()
        }

        if loss_fn is not None:
            result["loss"] = np.round(sum(losses) / len(losses), 5).item()

        return result

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(eval_every={self.eval_every}" + \
            f", n_classes={self.n_classes})[accuracy, precision, recall, f1]"

    def __repr__(self) -> str:
        return str(self)



from __future__ import annotations

import sys

import pytest
import torch
from torch.nn import Linear

sys.path.append(".")
sys.path.append("..")

from fluke import DDict  # NOQA
from fluke import FlukeENV  # NOQA
from fluke.client import Client  # NOQA
from fluke.comm import Channel  # NOQA
from fluke.data import FastDataLoader  # NOQA
from fluke.evaluation import ClassificationEval  # NOQA
from fluke.server import Server  # NOQA
from fluke.utils import OptimizerConfigurator, ServerObserver  # NOQA


def test_server():

    class Observer(ServerObserver):

        def start_round(self, round, global_model):
            assert round == 1
            assert global_model is not None

        def end_round(self,
                      round):
            assert round == 1

        def selected_clients(self, round, clients):
            assert len(clients) == 2
            assert round == 1

        def server_evaluation(self, round, type, evals, **kwargs) -> None:
            assert round == 1
            assert type == "global"
            assert "accuracy" in evals

        def finished(self, round):
            assert round == 1 or round == 2

    class Model(Linear):
        def __init__(self):
            super().__init__(10, 2)

    def target_function(x):
        return 0 if x[:7].sum() < 2.5 else 1

    FlukeENV().set_inmemory(True)
    Xtr = [torch.rand((100, 10)), torch.rand((100, 10))]
    ytr = [torch.tensor([target_function(x) for x in Xtr[0]]),
           torch.tensor([target_function(x) for x in Xtr[1]])]
    Xte = torch.rand((100, 10))
    yte = torch.tensor([target_function(x) for x in Xte])

    ftdl_client = [FastDataLoader(Xtr[i], ytr[i], num_labels=2, batch_size=10, shuffle=True)
                   for i in range(2)]
    ftdl_server = FastDataLoader(Xte, yte, num_labels=2, batch_size=10, shuffle=False)

    cfg = OptimizerConfigurator(optimizer_cfg=DDict(name=torch.optim.SGD, lr=0.1, momentum=0.9))
    clients = [Client(index=i,
                      train_set=ftdl_client[i],
                      test_set=ftdl_client[i] if i == 0 else None,
                      optimizer_cfg=cfg,
                      loss_fn=torch.nn.CrossEntropyLoss(),
                      local_epochs=3)
               for i in range(2)]

    server = Server(clients=clients,
                    model=Model(),
                    test_set=ftdl_server,
                    weighted=True)

    assert server.clients == clients
    assert isinstance(server.model, Model)
    assert server.test_set == ftdl_server
    assert server.hyper_params.weighted
    assert isinstance(server.channel, Channel)
    assert server.rounds == 0
    assert server.has_test
    assert server.has_model

    for c in clients:
        c.set_server(server)
        assert c.server == server
        assert c.channel == server.channel

    evaluator = ClassificationEval(1, 2)
    FlukeENV().set_evaluator(evaluator)
    obs = Observer()
    server.attach(obs)
    ev0 = server.evaluate(evaluator, server.test_set)
    server.fit(1, 1)
    ev1 = server.evaluate(evaluator, server.test_set)

    # assert ev0["loss"] >= ev1["loss"]
    assert server.rounds == 1
    assert str(server) == "Server(weighted=True, lr=1.0)"

    server.detach(obs)
    server.hyper_params.weighted = False
    try:
        server.fit(2, 0.5)
    except Exception:
        pytest.fail("Unexpected exception!")

    for c in clients:
        c.send_model()

    gen_models = server.receive_client_models(clients, state_dict=False)
    cmodels = list(gen_models)
    assert len(cmodels) == 2
    assert isinstance(cmodels[0], Model)

    server.test_set = None
    assert not server.evaluate(evaluator, server.test_set)

    server.broadcast_model(clients)

    for c in clients:
        m = c.channel.receive(c, server, "model")
        assert id(m) != id(server.model)
        assert m is not server.model

    assert str(server) == "Server(weighted=False, lr=1.0)"
    assert str(server) == repr(server)

    server.save("tmp/server")
    new_server = Server(clients=clients,
                        model=Model(),
                        test_set=ftdl_server,
                        weighted=False)
    new_server.load("tmp/server")
    assert str(new_server) == str(server)
    assert new_server.rounds == server.rounds
    assert new_server._participants == server._participants


if __name__ == "__main__":
    test_server()
    # 98% coverage for server.py



"""Implementation of the APFL [APFL20]_ algorithm.

References:
    .. [APFL20] Yuyang Deng, Mohammad Mahdi Kamani, and Mehrdad Mahdavi. Adaptive Personalized
       Federated Learning. In arXiv (2020). URL: https://arxiv.org/abs/2003.13461
"""
import sys
from typing import Any, Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client, PFLClient  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import merge_models  # NOQA

__all__ = [
    "APFLClient",
    "APFLServer",
    "APFL"
]


class APFLClient(PFLClient):

    def __init__(self,
                 index: int,
                 model: torch.nn.Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 lam: float = 0.25,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, model=model, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(lam=lam)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)

        self.model.train()
        self.personalized_model.train()

        self.model.to(self.device)
        self.personalized_model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        if self.pers_optimizer is None:
            self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.personalized_model)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)

                # Global
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()

                # Local
                self.pers_optimizer.zero_grad()
                y_hat = merge_models(self.model, self.personalized_model, self.hyper_params.lam)(X)
                local_loss = self.hyper_params.loss_fn(y_hat, y)
                local_loss.backward()
                self._clip_grads(self.personalized_model)
                self.pers_optimizer.step()

            self.scheduler.step()
            self.pers_scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        self.personalized_model.cpu()
        clear_cuda_cache()

        self.personalized_model = merge_models(
            self.model, self.personalized_model, self.hyper_params.lam)

        return running_loss


class APFLServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 tau: int = 3,
                 **kwargs: dict[str, Any]):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.hyper_params.update(tau=tau)

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        """Aggregate the models of the eligible clients every `hyper_params.tau` rounds.

        Args:
            eligible (Iterable[Client]): The clients that are eligible to participate in the
                aggregation.
            client_models (Iterable[Module]): The models of the clients to aggregate.
        """
        if self.rounds % self.hyper_params.tau != 0:
            # Ignore the sent models and clear the channel's cache
            self.channel.clear(self)
        else:
            super().aggregate(eligible, client_models)


class APFL(PersonalizedFL):

    def get_client_class(self) -> PFLClient:
        return APFLClient

    def get_server_class(self) -> Server:
        return APFLServer



"""Implementation of the FedProto [FedProto22]_ algorithm.
References:
    .. [FedProto22] Yue Tan, Guodong Long, Lu Liu, Tianyi Zhou, Qinghua Lu, Jing Jiang, Chengqi
       Zhang. FedProto: Federated Prototype Learning across Heterogeneous Clients. In AAAI (2022).
       URL: https://arxiv.org/abs/2105.00243
"""
import sys
from collections import defaultdict
from copy import deepcopy
from typing import Any, Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..evaluation import Evaluator  # NOQA
from ..nets import EncoderHeadNet  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache, get_model  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "FedProtoModel",
    "FedProtoClient",
    "FedProtoServer",
    "FedProto"
]


class FedProtoModel(Module):
    def __init__(self,
                 model: EncoderHeadNet,
                 prototypes: dict[int, torch.Tensor],
                 device: torch.device):
        super().__init__()
        self.model: EncoderHeadNet = model
        self.prototypes: dict[int, torch.Tensor] = prototypes
        self.num_classes: int = len(prototypes)
        self.device: torch.device = device

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        mse_loss = torch.nn.MSELoss()
        Z = self.model.encoder(x)
        output = float('inf') * torch.ones(x.shape[0], self.num_classes).to(self.device)
        for i, r in enumerate(Z):
            for j, proto in self.prototypes.items():
                # CHECKME: is the following lines necessary?
                if proto is not None and type(proto) is not type([]):
                    output[i, j] = mse_loss(r, proto)
                else:
                    output[i, j] = float('inf')

        # Return the negative of the distance so
        # to compute the argmax to get the closest prototype
        return -output


class FedProtoClient(Client):

    def __init__(self,
                 index: int,
                 model: Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 n_protos: int,
                 lam: float,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        fine_tuning_epochs = fine_tuning_epochs if fine_tuning_epochs else local_epochs
        super().__init__(index=index, model=model, train_set=train_set,
                         test_set=test_set, optimizer_cfg=optimizer_cfg, loss_fn=loss_fn,
                         local_epochs=local_epochs, fine_tuning_epochs=fine_tuning_epochs,
                         clipping=clipping, **kwargs)
        self.hyper_params.update(
            n_protos=n_protos,
            lam=lam
        )
        if isinstance(model, str):
            model = get_model(model)
        self.model = model
        self.prototypes = {i: None for i in range(self.hyper_params.n_protos)}
        self.global_protos = None

    def receive_model(self) -> None:
        msg = self.channel.receive(self, self.server, msg_type="model")
        self.global_protos = msg.payload

    def send_model(self):
        self.channel.send(Message(self.prototypes, "model", self, inmemory=True), self.server)

    def _update_protos(self, protos: Iterable[torch.Tensor]) -> None:
        for label, prts in protos.items():
            self.prototypes[label] = torch.sum(torch.vstack(prts), dim=0) / len(prts)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs
                       else self.hyper_params.local_epochs)

        self.model.train()
        self.model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        mse_loss = torch.nn.MSELoss()
        protos = defaultdict(list)
        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                Z = self.model.encoder(X)
                y_hat = self.model.head(Z)
                loss = self.hyper_params.loss_fn(y_hat, y)

                if self.server.rounds > 0:  # this is actually illegal in fluke :)
                    proto_new = deepcopy(Z.detach())
                    for i, yy in enumerate(y):
                        y_c = yy.item()
                        if self.global_protos[y_c] is not None:
                            proto_new[i, :] = self.global_protos[y_c].data
                        else:
                            proto_new[i, :] = torch.zeros_like(proto_new[i, :])
                    loss += self.hyper_params.lam * mse_loss(proto_new, Z)

                for i, yy in enumerate(y):
                    y_c = yy.item()
                    protos[y_c].append(Z[i, :].detach().data)

                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        clear_cuda_cache()
        self._update_protos(protos)
        return running_loss

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        if test_set is not None and self.prototypes[0] is not None:
            model = FedProtoModel(self.model, self.prototypes, self.device)
            return evaluator.evaluate(self._last_round, model, test_set, device=self.device)
        return {}

    def finalize(self) -> None:
        self._load_from_cache()
        self.fit(self.hyper_params.fine_tuning_epochs)
        self._save_to_cache()


class FedProtoServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = True,
                 n_protos: int = 10):
        super().__init__(model=None, test_set=None, clients=clients, weighted=weighted)
        self.hyper_params.update(n_protos=n_protos)
        self.prototypes = [None for _ in range(self.hyper_params.n_protos)]

    def broadcast_model(self, eligible: Iterable[Client]) -> None:
        # This function broadcasts the prototypes to the clients
        self.channel.broadcast(Message(self.prototypes, "model", self), eligible)

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        # Recieve models from clients, i.e., the prototypes
        clients_protos = client_models

        # Group by label
        label_protos = {i: [protos[i] for protos in clients_protos if protos[i] is not None]
                        for i in range(self.hyper_params.n_protos)}

        # Aggregate prototypes
        for label, protos in label_protos.items():
            if protos:
                self.prototypes[label] = torch.sum(torch.stack(protos), dim=0) / len(protos)


class FedProto(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedProtoClient

    def get_server_class(self) -> Server:
        return FedProtoServer



import json
import sys
import tempfile
from unittest.mock import patch

import pytest
import torch
from torch.nn import CrossEntropyLoss, Linear
from torch.optim import SGD
from torch.optim.lr_scheduler import StepLR

sys.path.append(".")
sys.path.append("..")


from fluke import DDict, FlukeENV  # NOQA
from fluke.algorithms import CentralizedFL  # NOQA
from fluke.client import Client  # NOQA
from fluke.comm import Message  # NOQA
from fluke.data import DataSplitter  # NOQA
from fluke.data.datasets import Datasets  # NOQA
from fluke.nets import MNIST_2NN, VGG9, FedBN_CNN, Shakespeare_LSTM  # NOQA
from fluke.server import Server  # NOQA
from fluke.utils import (ClientObserver, Configuration,  # NOQA
                         OptimizerConfigurator, ServerObserver, clear_cuda_cache,
                         get_class_from_qualified_name, get_class_from_str,
                         get_full_classname, get_loss, get_model, get_optimizer, bytes2human,
                         get_scheduler, import_module_from_str, memory_usage,
                         plot_distribution, retrieve_obj, cache_obj, flatten_dict)
from fluke.utils.log import Log, get_logger  # NOQA
from fluke.utils.model import (STATE_DICT_KEYS_TO_IGNORE,  # NOQA
                               AllLayerOutputModel, MMMixin,
                               batch_norm_to_group_norm, check_model_fit_mem,
                               diff_model, flatten_parameters,
                               get_global_model_dict, get_local_model_dict, get_trainable_keys,
                               merge_models, mix_networks, get_activation_size,
                               safe_load_state_dict, set_lambda_model, state_dict_zero_like)


def test_optimcfg():
    opt_cfg = OptimizerConfigurator(
        optimizer_cfg=dict(name=SGD,
                           lr=0.1,
                           momentum=0.9),
        scheduler_cfg=dict(
            step_size=1,
            gamma=.1
        ),
    )

    assert opt_cfg.optimizer == SGD
    assert opt_cfg.scheduler_cfg == {
        "step_size": 1,
        "gamma": 0.1
    }

    assert opt_cfg.optimizer_cfg == {
        "lr": 0.1,
        "momentum": 0.9
    }

    opt_cfg = OptimizerConfigurator(
        optimizer_cfg=DDict(name="SGD",
                            lr=0.1,
                            momentum=0.9),
        scheduler_cfg=DDict(
            name=StepLR,
            step_size=1,
            gamma=.1
        ),
    )

    assert opt_cfg.optimizer == SGD
    assert opt_cfg.scheduler == StepLR

    assert opt_cfg.optimizer_cfg == {
        "lr": 0.1,
        "momentum": 0.9
    }

    opt, sch = opt_cfg(Linear(10, 10))

    assert isinstance(opt, SGD)
    assert isinstance(sch, StepLR)
    assert opt.defaults["lr"] == 0.1
    assert opt.defaults["momentum"] == 0.9
    assert sch.step_size == 1
    assert sch.gamma == 0.1
    assert str(opt_cfg) == "OptCfg(SGD, lr=0.1, momentum=0.9, StepLR(step_size=1, gamma=0.1))"
    assert str(opt_cfg) == opt_cfg.__repr__()

    with pytest.raises(ValueError):
        opt_cfg = OptimizerConfigurator(
            optimizer_cfg=dict(name=1,
                               lr=0.1,
                               momentum=0.9))
    with pytest.raises(ValueError):
        opt_cfg = OptimizerConfigurator(
            optimizer_cfg=DDict(name=SGD,
                                lr=0.1,
                                momentum=0.9),
            scheduler_cfg=[DDict(
                step_size=1,
                gamma=.1
            )]
        )

    with pytest.raises(ValueError):
        opt_cfg = OptimizerConfigurator(
            optimizer_cfg=[DDict(name=SGD,
                                 lr=0.1,
                                 momentum=0.9)],
            scheduler_cfg=DDict(
                step_size=1,
                gamma=.1
            )
        )

    with pytest.raises(ValueError):
        opt_cfg = OptimizerConfigurator(
            optimizer_cfg=DDict(name=SGD,
                                lr=0.1,
                                momentum=0.9),
            scheduler_cfg=[DDict(
                name="Pippo",
                step_size=1,
                gamma=.1
            )]
        )


def test_functions():
    # try:
    client_module = import_module_from_str("fluke.client")
    client_class = get_class_from_str("fluke.client", "Client")
    assert client_class == Client
    model = get_model("MNIST_2NN")
    assert model.__class__.__name__ == "MNIST_2NN"
    model2 = get_model("fluke.nets.MNIST_2NN")
    assert model2.__class__.__name__ == "MNIST_2NN"
    linear = get_class_from_qualified_name("torch.nn.Linear")
    assert linear == Linear
    full_linear = get_full_classname(Linear)
    optim = get_optimizer("SGD")
    assert optim == SGD
    loss = get_loss("CrossEntropyLoss")
    assert isinstance(loss, CrossEntropyLoss)
    scheduler = get_scheduler("StepLR")
    assert scheduler == StepLR
    logger = get_logger("Log")
    assert isinstance(logger, Log)
    clear_cuda_cache()
    clear_cuda_cache(True)
    # except Exception:
    #     pytest.fail("Unexpected error!")

    rss, vms, cud = memory_usage()
    assert rss >= 0
    assert vms >= 0
    assert cud >= 0
    if not torch.cuda.is_available():
        assert cud == 0

    assert client_module.__name__ == "fluke.client"
    assert client_class == Client
    assert model.__class__.__name__ == "MNIST_2NN"
    assert model2.__class__.__name__ == "MNIST_2NN"
    assert linear == Linear
    assert full_linear == "torch.nn.modules.linear.Linear"
    assert isinstance(loss, CrossEntropyLoss)
    assert scheduler == StepLR
    assert isinstance(logger, Log)

    model3 = torch.nn.Sequential(
        torch.nn.Conv2d(3, 6, 3, 1, 1),
        torch.nn.BatchNorm2d(6),
        torch.nn.ReLU()
    )

    model4 = torch.nn.Sequential(
        torch.nn.Conv2d(3, 6, 3, 1, 1),
        torch.nn.BatchNorm2d(6),
        torch.nn.ReLU()
    )

    batch = torch.randn(1, 3, 28, 28)
    model4(batch)

    assert bytes2human(100) == "100 B"

    sd = state_dict_zero_like(model4.state_dict())
    assert list(sd.keys()) == list(model4.state_dict().keys())
    for k in sd.keys():
        assert torch.all(sd[k] == 0)

    prev_state_dict = model3.state_dict()
    safe_load_state_dict(model3, model4.state_dict())
    for k in model3.state_dict().keys():
        if k.endswith(STATE_DICT_KEYS_TO_IGNORE):
            assert torch.all(prev_state_dict[k] == model3.state_dict()[k])
        else:
            assert torch.all(model3.state_dict()[k] == model4.state_dict()[k])

    FlukeENV().set_inmemory(False)
    FlukeENV().open_cache("test_utils")
    cache_obj(model3, "test")
    load_model3 = retrieve_obj("test")

    server = Server(model3, None, [], None, None)
    cache_obj(model3, "model", server)
    load_cmodel = retrieve_obj("model", server)

    assert load_model3 is not None
    assert load_cmodel is not None
    assert isinstance(load_model3, torch.nn.Sequential)
    assert isinstance(load_cmodel, torch.nn.Sequential)

    tkeys = get_trainable_keys(model3)
    assert len(tkeys) == 4
    assert list(tkeys) == ["0.weight", "0.bias", "1.weight", "1.bias"]

    FlukeENV().get_cache().cleanup()
    FlukeENV().close_cache()
    cache_obj(None, "test")

    for k in model3.state_dict().keys():
        assert torch.all(model3.state_dict()[k] == load_model3.state_dict()[k])

    # client = Client(1, None, None, None, None)
    # client.model = model3
    # cache_model(client, ["model"])
    # load_model3 = retrieve_model(client, ["model"])["model"]

    for k in model3.state_dict().keys():
        assert torch.all(model3.state_dict()[k] == load_model3.state_dict()[k])

    d = {"a": 1, "b": 2, "c": {"d": 3}}
    dflat = flatten_dict(d)
    assert dflat == {"a": 1, "b": 2, "c.d": 3}


def test_configuration():

    cfg = dict({
        'protocol': {
            'n_clients': 100,
            'n_rounds': 50,
            'eligible_perc': 0.1
        },
        'data': {
            'dataset': {
                'name': 'mnist'
            },
            'distribution': {
                'name': "iid"
            },
            'client_split': 0.1,
            'sampling_perc': 1
        },
        'save': {
            'path': 'temp',
            'save_every': 1,
            'global_only': True
        },
        'exp': {
            'seed': 42,
            'average': 'micro',
            'device': 'cpu'
        },
        'logger': {
            'name': 'local'
        }
    })
    cfg_alg = dict({
        'name': 'fluke.algorithms.fedavg.FedAVG',
        'hyperparameters': {
            'server': {
                'weighted': True
            },
            'client': {
                'batch_size': 10,
                'local_epochs': 5,
                'loss': 'CrossEntropyLoss',
                'optimizer': {
                    'lr': 0.01,
                    'momentum': 0.9,
                    'weight_decay': 0.0001
                },
                'scheduler': {
                    'step_size': 1,
                    'gamma': 1
                }
            },
            'model': 'MNIST_2NN'
        }
    })

    temp_cfg = tempfile.NamedTemporaryFile(mode="w")
    temp_cfg_alg = tempfile.NamedTemporaryFile(mode="w")
    json.dump(cfg, open(temp_cfg.name, "w"))
    json.dump(cfg_alg, open(temp_cfg_alg.name, "w"))

    try:
        conf = Configuration(temp_cfg.name, temp_cfg_alg.name)
    except Exception:
        pytest.fail("Unexpected error!")

    cfg["exp"]["seed"] = [42, 50, 133]
    temp_cfg = tempfile.NamedTemporaryFile(mode="w")
    temp_cfg_alg = tempfile.NamedTemporaryFile(mode="w")
    json.dump(cfg, open(temp_cfg.name, "w"))
    json.dump(cfg_alg, open(temp_cfg_alg.name, "w"))

    gencfg = Configuration.sweep(temp_cfg.name, temp_cfg_alg.name)
    assert len(list(gencfg)) == 3

    cfg["exp"]["seed"] = 42

    assert conf.protocol.n_clients == 100
    # assert conf.data.dataset.name == "mnist"
    assert conf.exp.seed == 42
    # assert conf.logger.name == "local"
    assert conf.client.local_epochs == 5
    assert conf.server.weighted
    assert conf.model == "MNIST_2NN"

    assert str(conf) == "fluke.algorithms.fedavg.FedAVG_data(mnist, iid())_proto(C100, R50, E0.1)_seed(42)"
    assert conf.__repr__() == str(conf)

    cfg = dict({"protocol": {}, "data": {}, "exp": {}, "logger": {}})
    cfg_alg = dict({"name": "fluke.algorithms.fedavg.FedAVG", "hyperparameters": {
        "server": {}, "client": {}, "model": "MNIST_2NN"}})

    temp_cfg = tempfile.NamedTemporaryFile(mode="w")
    temp_cfg_alg = tempfile.NamedTemporaryFile(mode="w")
    json.dump(cfg, open(temp_cfg.name, "w"))
    json.dump(cfg_alg, open(temp_cfg_alg.name, "w"))

    with pytest.raises(ValueError):
        conf = Configuration(temp_cfg.name, temp_cfg_alg.name)

    cfg["logger"]["name"] = "WandBLog"
    cfg_ = cfg.copy()
    cfg_["method"] = cfg_alg
    with pytest.raises(ValueError):
        Configuration.from_ddict(DDict(cfg_))


def test_log():
    log = Log()
    log.init(test="hello")

    try:
        log.comm_costs[0] = 1  # for testing
        log.start_round(1, None)
        log.comm_costs[0] = 0  # for testing
        log.selected_clients(1, [1, 2, 3])
        log.message_received(Message("test", "test", None))
        log.server_evaluation(1, "global", {"accuracy": 1})
        log.client_evaluation(1, 1, 'pre-fit', {"accuracy": 0.6})
        log.end_round(1)
        log.finished(1)
        temp = tempfile.NamedTemporaryFile(mode="w")
        log.save(temp.name)
    except Exception:
        pytest.fail("Unexpected error!")

    with open(temp.name, "r") as f:
        data = dict(json.load(f))
        assert data == {'perf_global': {'1': {'accuracy': 1}}, 'comm_costs': {
            '0': 0, '1': 4}, 'perf_locals': {}, 'perf_prefit': {'1': {'accuracy': 0.6}},
            'perf_postfit': {'1': {'accuracy': 0.6}}}

    assert log.global_eval == {1: {"accuracy": 1}}
    assert log.locals_eval == {}
    assert log.prefit_eval == {1: {1: {"accuracy": 0.6}}}
    assert log.postfit_eval == {}
    assert log.locals_eval_summary == {}
    assert log.prefit_eval_summary == {1: {"accuracy": 0.6}}
    assert log.postfit_eval_summary == {1: {'accuracy': 0.6}}
    assert log.comm_costs == {0: 0, 1: 4}
    assert log.current_round == 1


# def test_wandb_log():
#     log2 = WandBLog()
#     log2.init()
#     try:
#         log2.comm_costs[0] = 1  # for testing
#         log2.start_round(1, None)
#         log2.comm_costs[0] = 0  # for testing
#         log2.selected_clients(1, [1, 2, 3])
#         log2.message_received(Message("test", "test", None))
#         log2.error("test")
#         log2.end_round(1, {"accuracy": 1}, [{"accuracy": 0.7}, {"accuracy": 0.5}])
#         log2.finished([{"accuracy": 0.7}, {"accuracy": 0.5}, {"accuracy": 0.6}])
#         temp = tempfile.NamedTemporaryFile(mode="w")
#         log2.save(temp.name)
#     except Exception:
#         pytest.fail("Unexpected error!")

#     with open(temp.name, "r") as f:
#         data = dict(json.load(f))
#         assert data == {'perf_global': {'1': {'accuracy': 1}}, 'comm_costs': {
#             '0': 0, '1': 4}, 'perf_local': {'1': {'accuracy': 0.6}, '2': {'accuracy': 0.6}}}

#     assert log2.history[1] == {"accuracy": 1}
#     assert log2.client_history[1] == {"accuracy": 0.6}
#     assert log2.comm_costs[1] == 4


def test_models():
    model1 = Linear(2, 1)
    model1.weight.data.fill_(1)
    model1.bias.data.fill_(1)

    model2 = Linear(2, 1)
    model2.weight.data.fill_(2)
    model2.bias.data.fill_(2)

    model3 = merge_models(model1, model2, 0.5)
    assert model3.weight.data[0, 0] == 1.5
    assert model3.weight.data[0, 1] == 1.5
    assert model3.bias.data[0] == 1.5

    model4 = merge_models(model1, model2, 0.75)
    assert model4.weight.data[0, 0] == 1.75
    assert model4.weight.data[0, 1] == 1.75
    assert model4.bias.data[0] == 1.75

    diffdict = diff_model(model1.state_dict(), model2.state_dict())
    assert diffdict["weight"].data[0, 0] == -1.0
    assert diffdict["weight"].data[0, 1] == -1.0
    assert diffdict["bias"].data[0] == -1.0

    model = FedBN_CNN()
    model_gn = batch_norm_to_group_norm(model)

    for n1, p1 in model.named_parameters():
        for n2, p2 in model_gn.named_parameters():
            if n1 == n2 and isinstance(p1, torch.nn.BatchNorm2d):
                assert isinstance(p2, torch.nn.GroupNorm)


def test_mixing():
    mixin = MMMixin()
    mixin.set_lambda(0.5)
    assert mixin.get_lambda() == 0.5

    model1 = Linear(2, 1)
    model1.weight.data.fill_(1)
    model1.bias.data.fill_(1)

    model2 = Linear(2, 1)
    model2.weight.data.fill_(2)
    model2.bias.data.fill_(2)
    mixed = mix_networks(model1, model2, 0.5)

    x = torch.FloatTensor([[1, 1]])
    y = mixed(x)
    assert y[0, 0] == 4.5

    assert hasattr(mixed, "lam")
    assert mixed.lam == 0.5
    assert mixed.get_lambda() == 0.5
    assert diff_model(get_local_model_dict(mixed), model2.state_dict())["weight"].data[0, 0] == 0.0
    assert diff_model(get_global_model_dict(mixed), model1.state_dict())["weight"].data[0, 0] == 0.0

    set_lambda_model(mixed, 0.3)

    assert mixed.lam == 0.3
    assert mixed.get_lambda() == 0.3

    weights = mixed.get_weight()
    assert weights[0].data[0, 0] == 0.7 + 0.3 * 2
    assert weights[0].data[0, 1] == 0.7 + 0.3 * 2
    assert weights[1].data[0] == 0.7 + 0.3 * 2

    model1 = MNIST_2NN()
    model2 = MNIST_2NN()
    mixed = mix_networks(model1, model2, 0.2)

    assert mixed.get_lambda() == 0.2

    model1 = VGG9()
    model2 = VGG9()
    mixed = mix_networks(model1, model2, 0.3)
    assert mixed.get_lambda() == 0.3

    x = torch.randn(1, 1, 28, 28)
    mixed(x)

    model1 = Shakespeare_LSTM()
    model2 = Shakespeare_LSTM()
    mixed = mix_networks(model1, model2, 0.4)
    assert mixed.get_lambda() == 0.4

    x = torch.randint(0, 100, (1, 10))
    mixed(x)

    class TestNet(torch.nn.Module):
        def __init__(self):
            super(TestNet, self).__init__()

            # Implement the sequential module for feature extraction
            self.features = torch.nn.Sequential(
                torch.nn.Conv2d(in_channels=1, out_channels=10, kernel_size=3, stride=1, padding=1),
                torch.nn.MaxPool2d(2, 2), torch.nn.ReLU(inplace=True), torch.nn.BatchNorm2d(10),
                torch.nn.Conv2d(in_channels=10, out_channels=20,
                                kernel_size=3, stride=1, padding=1),
                torch.nn.MaxPool2d(2, 2), torch.nn.ReLU(inplace=True), torch.nn.BatchNorm2d(20)
            )

            # Implement the fully connected layer for classification
            self.fc = torch.nn.Linear(in_features=20 * 7 * 7, out_features=10)

        def forward(self, x):
            x = self.features(x)
            x = x.view(x.size(0), -1)
            x = self.fc(x)
            return x

    model1 = TestNet()
    model2 = TestNet()
    mixed = mix_networks(model1, model2, 0.3)
    assert mixed.get_lambda() == 0.3

    x = torch.randn(1, 1, 28, 28)
    mixed(x)


def test_serverobs():
    sobs = ServerObserver()
    sobs.start_round(1, None)
    sobs.server_evaluation(1, "global", {"accuracy": 1})
    sobs.end_round(1)
    sobs.finished(1)
    sobs.selected_clients(1, [1, 2, 3])


def test_clientobs():
    sobs = ClientObserver()
    sobs.start_fit(1, 17, None)
    sobs.client_evaluation(1, 17, "pre-fit", {"accuracy": 1})
    sobs.end_fit(1, 17, None, 0.1)


# @patch("matplotlib.pyplot.show")
# def test_plot_dist(mock_show):
#     hparams = DDict(
#         # model="fluke.nets.MNIST_2NN",
#         model=MNIST_2NN(),
#         client=DDict(batch_size=32,
#                      local_epochs=1,
#                      loss=CrossEntropyLoss,
#                      optimizer=DDict(
#                          lr=0.1,
#                          momentum=0.9),
#                      scheduler=DDict(
#                          step_size=1,
#                          gamma=0.1)
#                      ),
#         server=DDict(weighted=True)
#     )
#     mnist = Datasets.MNIST("../data")
#     splitter = DataSplitter(mnist, client_split=0.1)
#     fl = CentralizedFL(10, splitter, hparams)
#     plot_distribution(fl.clients, "ball")
#     plot_distribution(fl.clients, "bar")
#     plot_distribution(fl.clients, "mat")
#     FlukeENV().close_cache()


@patch("matplotlib.pyplot.show")
def test_plot_dist_ball(mock_show):
    hparams = DDict(
        # model="fluke.nets.MNIST_2NN",
        model=MNIST_2NN(),
        client=DDict(batch_size=32,
                     local_epochs=1,
                     loss=CrossEntropyLoss,
                     optimizer=DDict(
                         lr=0.1,
                         momentum=0.9),
                     scheduler=DDict(
                         step_size=1,
                         gamma=0.1)
                     ),
        server=DDict(weighted=True)
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist, client_split=0.1)
    fl = CentralizedFL(10, splitter, hparams)
    plot_distribution(fl.clients, type="ball")
    FlukeENV().close_cache()


@patch("matplotlib.pyplot.show")
def test_plot_dist_bar(mock_show):
    hparams = DDict(
        # model="fluke.nets.MNIST_2NN",
        model=MNIST_2NN(),
        client=DDict(batch_size=32,
                     local_epochs=1,
                     loss=CrossEntropyLoss,
                     optimizer=DDict(
                         lr=0.1,
                         momentum=0.9),
                     scheduler=DDict(
                         step_size=1,
                         gamma=0.1)
                     ),
        server=DDict(weighted=True)
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist, client_split=0.1)
    fl = CentralizedFL(10, splitter, hparams)
    plot_distribution(fl.clients, type="bar")
    FlukeENV().close_cache()


@patch("matplotlib.pyplot.show")
def test_plot_dist_mat(mock_show):
    hparams = DDict(
        # model="fluke.nets.MNIST_2NN",
        model=MNIST_2NN(),
        client=DDict(batch_size=32,
                     local_epochs=1,
                     loss=CrossEntropyLoss,
                     optimizer=DDict(
                         lr=0.1,
                         momentum=0.9),
                     scheduler=DDict(
                         step_size=1,
                         gamma=0.1)
                     ),
        server=DDict(weighted=True)
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist, client_split=0.1)
    fl = CentralizedFL(10, splitter, hparams)
    plot_distribution(fl.clients, type="mat")
    FlukeENV().close_cache()


def test_check_mem():
    net = MNIST_2NN()

    if torch.backends.mps.is_available():
        assert check_model_fit_mem(net, (28 * 28,), 100, "mps", True)

    if torch.cuda.is_available():
        assert check_model_fit_mem(net, (28 * 28,), 100, "cuda")


def test_get_activation_size():
    net = MNIST_2NN()
    x = torch.randn(1, 28 * 28)
    assert 10 == get_activation_size(net, None)
    assert 10 == get_activation_size(net, x)

    net = FedBN_CNN()
    x = torch.randn(1, 1, 28, 28)
    with pytest.raises(ValueError):
        get_activation_size(net.encoder, None)
    assert 10 == get_activation_size(net, x)


def test_alllayeroutput():
    net = MNIST_2NN()
    all_out = AllLayerOutputModel(net)
    x = torch.randn(1, 28 * 28)
    all_out(x)
    assert "_encoder.fc1" in all_out.activations_in
    assert "_encoder.fc2" in all_out.activations_in
    assert "_head.fc3" in all_out.activations_in
    assert "_encoder.fc1" in all_out.activations_out
    assert "_encoder.fc2" in all_out.activations_out
    assert "_head.fc3" in all_out.activations_out

    all_out.deactivate()
    all_out(x)
    assert all_out.activations_in == {}
    assert all_out.activations_out == {}


def test_flatten():
    net = MNIST_2NN()
    W = flatten_parameters(net)
    assert W.shape[0] == 178110


if __name__ == "__main__":
    test_optimcfg()
    test_functions()
    test_configuration()
    test_log()
    # test_wandb_log()
    test_models()
    test_mixing()
    test_serverobs()
    test_clientobs()
    # test_plot_dist()
    test_check_mem()
    test_alllayeroutput()

    # 91% coverage utils.__init__
    # 95% coverage utils.model



"""Implementation of the Federated Averaging [FedAVG17]_ algorithm.

References:
    .. [FedAVG17] H. Brendan McMahan, Eider Moore, Daniel Ramage, Seth Hampson, and Blaise Aguera y
       Arcas. Communication-efficient learning of deep networks from decentralized data.
       In AISTATS (2017). URL: https://arxiv.org/abs/1602.05629

"""
import sys

sys.path.append(".")
sys.path.append("..")

from . import CentralizedFL  # NOQA

__all__ = [
    "FedAVG"
]


class FedAVG(CentralizedFL):
    pass
